from __future__ import annotations

from http import HTTPStatus
from string import Formatter
from typing import TYPE_CHECKING, Any, ClassVar, Self

from graphql import GraphQLError

from undine.typing import GraphQLWebSocketCloseCode, UndineErrorCodes

if TYPE_CHECKING:
    from collections.abc import Collection, Generator, Sequence

    from graphql import GraphQLErrorExtensions, Node, Source


class ErrorMessageFormatter(Formatter):
    """Formatter for error strings."""

    def format_field(self, value: Any, format_spec: str) -> str:
        from undine.utils.text import comma_sep_str, dotpath  # noqa: PLC0415

        if format_spec == "dotpath":
            return dotpath(value)
        if format_spec == "module":
            return value.__module__
        if format_spec == "name":
            return value.__name__
        if format_spec == "qualname":
            return value.__qualname__
        if format_spec == "type":
            return type(value).__name__
        if format_spec == "comma_sep_or":
            return comma_sep_str(value, last_sep="or", quote=True)
        if format_spec == "comma_sep_and":
            return comma_sep_str(value, last_sep="and", quote=True)

        return super().format_field(value, format_spec)


# Undine Errors


class UndineError(Exception):
    """Base class for all undine errors."""

    msg: ClassVar[str] = ""
    error_formatter = ErrorMessageFormatter()

    def __init__(
        self,
        msg: str = "",
        **kwargs: Any,
    ) -> None:
        msg = msg or self.msg
        if kwargs:
            msg = self.error_formatter.format(msg, **kwargs)
        super().__init__(msg)


class UndineErrorGroup(ExceptionGroup):
    """Base class for all undine exception groups."""

    msg: ClassVar[str] = ""
    error_formatter = ErrorMessageFormatter()

    def __new__(
        cls,
        errors: Sequence[Exception],
        *,
        msg: str = "",
        **kwargs: Any,
    ) -> Self:
        msg = msg or cls.msg
        if kwargs:
            msg = cls.error_formatter.format(msg, **kwargs)
        return super().__new__(UndineErrorGroup, msg, errors)  # type: ignore[return-value]

    def __init__(self, errors: Sequence[Exception], *, msg: str = "", **kwargs: Any) -> None:
        super().__init__(msg, errors)

    def flatten(self) -> Generator[Exception, None, None]:
        """Flattened the errors inside the `UndineErrorGroup`."""
        for error in self.exceptions:
            if isinstance(error, UndineErrorGroup):
                yield from error.flatten()
            elif isinstance(error, Exception):
                yield error


class BulkMutateNeedsImplementationError(UndineError):
    """Error raised if a MutationType '__bulk_mutate__' method is not implemented when its should be."""

    msg = "Must implement '{mutation_type}.__bulk_mutate__' to handle related inputs"


class DirectiveLocationError(UndineError):
    """Error raised if Directive is passed to a location it cannot be used in."""

    msg = "Directive {directive!r} is not allowed in {location.name!r}"


class EmptyFilterResult(UndineError):  # noqa: N818
    """Error that should be raised when using a filter should result in an empty queryset."""

    msg = "Filter result should be null."


class ExpressionMultipleOutputFieldError(UndineError):
    """Error raised if no output field cannot be determined for an expression."""

    msg = (
        "Could not determine an output field for expression {expr!r}. "
        "Got multiple possible output fields: {output_fields}."
    )


class ExpressionNoOutputFieldError(UndineError):
    """Error raised if no output field cannot be determined for an expression."""

    msg = (
        "Could not determine an output field for expression {expr!r}. "
        "No output field found from any source expressions."
    )


class FunctionSignatureParsingError(UndineError):
    """Error raised if a function is missing type annotations for its parameters."""

    msg = (
        "Type '{name}' is not defined in module '{func:module}'. "
        "Check if it's inside a `if TYPE_CHECKING` block or another class/function. "
        "The type needs to be available at the runtime so that the signature of '{func:qualname}' can be inspected."
    )


class InterfaceFieldDoesNotExistError(UndineError):
    """
    Error raised when a QueryType is trying to inherit a field from an InterfaceType,
    but the field doesn't exist on the Model.
    """

    msg = "Field '{field}' from interface '{interface:dotpath}' does not exist on Model '{model:dotpath}'."


class InterfaceFieldTypeMismatchError(UndineError):
    """
    Error raised when a QueryType is trying to inherit a field from an InterfaceType,
    but its converted GraphQL type doesn't match the GraphQL type converted from the matching Model field.
    """

    msg = (
        "Field '{field}' from interface '{interface:dotpath}' expects type '{output_type}' "
        "but Model field generated type '{field_type}'"
    )


class InvalidInputMutationTypeError(UndineError):
    """Error raised when trying to create an `Input` using a `MutationType` with a `kind` other than 'related'."""

    msg = (
        "MutationType '{ref:dotpath}' is a '{kind}' MutationType, "
        "but only 'related' MutationTypes can be used as Inputs on other MutationTypes."
    )


class InvalidDocstringParserError(UndineError):
    """Error raised when an invalid docstring parser is provided."""

    msg = "'{cls:dotpath}' does not implement 'DocstringParserProtocol'."


class InvalidEntrypointMutationTypeError(UndineError):
    """Error raised when trying to create an `Entrypoint` using a `MutationType` with an unknown `kind`."""

    msg = (
        "MutationType '{ref:dotpath}' is a '{kind}' MutationType, "
        "but only 'create', 'update', 'delete', or 'custom' MutationTypes can be used in Entrypoints."
    )


class MismatchingModelError(UndineError):
    """
    Error raised if provided model for `FilterSet` or `OrderSet`
    doesn't match model of the given `QueryType`.
    """

    msg = "'{name}' model '{given_model:dotpath}' does not match '{target}' model '{expected_model:dotpath}'."


class MissingCalculationReturnTypeError(UndineError):
    """Error raised if a calculation class doesn't define a return type."""

    msg = (
        "'{name}' must define the calculation return type using the Generic type argument: "
        "e.g. `class {name}(Calculation[int]):`"
    )


class MissingEntrypointRefError(UndineError):
    """Error raised when an entrypoint is missing a reference."""

    msg = "Entrypoint '{name}' in class '{cls:dotpath}' must have a reference."


class MissingFunctionAnnotationsError(UndineError):
    """Error raised if a function is missing type annotations for its parameters."""

    msg = "Missing type hints for parameters {missing:comma_sep_and} in function '{func:dotpath}'."


class MissingFunctionReturnTypeError(UndineError):
    """Error raised if a function does not contain a parameter to parse type from."""

    msg = "Missing type hint for return value in function '{func:dotpath}'."


class MissingDirectiveArgumentError(UndineError):
    """Error raised if a directive argument is missing."""

    msg = "Missing directive argument '{name}' for directive '{directive:dotpath}'."


class MissingDirectiveLocationsError(UndineError):
    """Error raised if no locations are provided to `Directive`."""

    msg = (
        "'{name}' is missing `locations` keyword argument in its class definition: "
        "e.g. `class {name}(Directive, locations=[DirectiveLocation.FIELD_DEFINITION])`."
    )


class MissingModelGenericError(UndineError):
    """Error raised if no model is provided to `QueryType`, `MutationType`, `FilterSet`, or `OrderSet`."""

    msg = "'{name}' is missing its generic types: `class {name}({cls}[MyModel])`."


class MissingUnionQueryTypeGenericError(UndineError):
    """Error raised if no models are provided to `UnionType`."""

    msg = "'{name}' is missing its generic types: `class {name}(UnionType[QueryType1, QueryType2])`."


class ModelFieldError(UndineError): ...


class ModelFieldDoesNotExistError(ModelFieldError):
    """Error raised if a field does not exist in the given model."""

    msg = "Field '{field}' does not exist in model '{model:dotpath}'."


class ModelFieldNotARelationError(ModelFieldError):
    """Error raised if a field is not a relation in the given model."""

    msg = "Field '{field}' is not a relation in model '{model:dotpath}'."


class ModelFieldNotARelationOfModelError(ModelFieldError):
    """Error raised if a field is not a relation in the given model."""

    msg = "Field '{field}' is not a relation from model '{model:dotpath}' to model '{related:dotpath}'."


class MutateNeedsImplementationError(UndineError):
    """Error raised if a MutationType '__mutate__' method is not implemented when its should be."""

    msg = "Must implement '{mutation_type}.__mutate__' to handle related inputs"


class MutationTypeKindCannotBeDeterminedError(UndineError):
    """Error raised if mutation type cannot determine its kind automatically."""

    msg = "Cannot determine mutation kind for MutationType '{name}'"


class NoFunctionParametersError(UndineError):
    """Error raised if a function does not contain a parameter to parse type from."""

    msg = "Function '{func:dotpath}' must have at least one argument."


class NotCompatibleWithError(UndineError):
    """Error raised if a given object is not compatible with some other object."""

    msg = "Cannot use '{obj:dotpath}' with '{other:dotpath}'"


class FunctionDispatcherError(UndineError):
    """Error raised for `FunctionDispatcher` errors."""


class FunctionDispatcherImplementationNotFoundError(FunctionDispatcherError):
    """Error raised when `FunctionDispatcher` cannot find an implementation for a given key."""

    msg = "'{name}' doesn't contain an implementation for {key!r} (type: {cls:dotpath})."


class FunctionDispatcherImproperLiteralError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with a Literal that has an invalid value.
    """

    msg = "Literal argument must be a string, integer, bytes, boolean, enum, or None, got {arg!r}."


class FunctionDispatcherNoArgumentAnnotationError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with a function that doesn't have a type hint for its first argument.
    """

    msg = (
        "Function '{func_name}' must have a type hint for its first argument so that it can be registered for '{name}'."
    )


class FunctionDispatcherNoArgumentsError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with a function that doesn't have any arguments.
    """

    msg = "Function '{func_name}' must have at least one argument so that it can be registered for '{name}'."


class FunctionDispatcherNonRuntimeProtocolError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with a protocol that hasn't been decorated with `@runtime_checkable`.
    """

    msg = "Protocol '{name}' is not a runtime checkable protocol."


class FunctionDispatcherRegistrationError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    that is something other than a function.
    """

    msg = "Can only register functions with '{name}'. Got {value!r}."


class FunctionDispatcherUnknownArgumentError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with an unknown argument type.
    """

    msg = "Unknown argument: {annotation!r}"


class QueryTypeRequiresSingleModelError(UndineError):
    """Error raised when a FilterSet or OrderSet with multiple models is added to a QueryType."""

    msg = "Cannot add a {kind} with multiple models to a QueryType"


class RegistryDuplicateError(UndineError):
    """Error raised if trying to register a value for the same key twice."""

    msg = "'{registry_name}' already contains a value for '{key}': '{value}'"


class RegistryMissingTypeError(UndineError):
    """Error raised when a Registry doesn't contain an entry for a given key."""

    msg = "'{registry_name}' doesn't contain an entry for '{key}'"


class UnexpectedDirectiveArgumentError(UndineError):
    """Error raised if a directive argument is unexpected."""

    msg = "Unexpected directive arguments for directive '{directive:dotpath}': {kwargs}."


class UnionTypeMultipleTypesError(FunctionDispatcherError):
    """
    Error raised when a trying to register an implementation for a `FunctionDispatcher`
    with a Union type that has more than one non-null type.
    """

    msg = "Union type must have a single non-null type argument, got {args}."


class UnionModelFieldMismatchError(UndineError):
    """
    Error raised when trying to create a Filter or an Order from a string reference,
    and the corresponding FilterSet or OrderSet contains multiple models,
    but the string references convert to different graphql types.
    """

    msg = (
        "'{ref}' is of type '{type_1}' when converted from model '{model_1:dotpath}' "
        "but of type '{type_2}' when converted from model '{model_2:dotpath}'. "
        "Cannot create a {kind} due to mismatching types."
    )


class UnionModelFieldDirectUsageError(UndineError):
    """
    Error raised when trying to create a Filter or an Order from a model field directly,
    but the corresponding FilterSet or OrderSet contains many models.
    """

    msg = "Cannot use model reference when {kind} defined for multiple models"


class UnionTypeModelsDifferentError(UndineError):
    """Error raised when a FilterSet or OrderSet is added to a UnionType with different models."""

    msg = "Cannot add a {kind} to a UnionType with different models"


class UnionTypeRequiresMultipleModelsError(UndineError):
    """Error raised when a FilterSet or OrderSet with a single model is added to a UnionType."""

    msg = "Cannot add a {kind} with a single model to a UnionType"


# GraphQL Errors


class GraphQLStatusError(GraphQLError):
    """Base error for GraphQL error in Undine."""

    msg: ClassVar[str] = ""
    status: ClassVar[int] = HTTPStatus.INTERNAL_SERVER_ERROR
    code: ClassVar[str | None] = None
    error_formatter = ErrorMessageFormatter()

    def __init__(
        self,
        message: str = "",
        *,
        status: int | None = None,
        code: str | None = None,
        nodes: Collection[Node] | Node | None = None,
        source: Source | None = None,
        positions: Collection[int] | None = None,
        path: Collection[str | int] | None = None,
        original_error: Exception | None = None,
        extensions: GraphQLErrorExtensions | None = None,
        **kwargs: Any,
    ) -> None:
        """
        Initialize the GraphQL Error with some extra information.

        :param message: A message describing the Error for debugging purposes.
        :param status: HTTP status code.
        :param code: Unique error code.
        :param nodes: A list of GraphQL AST Nodes corresponding to this error
        :param source: The source GraphQL document for the first location of this error.
        :param positions: A list of character offsets within the source GraphQL document which correspond
                          to this error.
        :param path: A list of field names and array indexes describing the JSON-path into the execution
                     response which corresponds to this error.
        :param original_error: The original error thrown from a field resolver during execution.
        :param extensions: Extension fields to add to the formatted error.
        """
        status = status or self.status
        code = code or self.code

        message = message or self.msg
        if kwargs:
            message = self.error_formatter.format(message, **kwargs)

        extensions = extensions or {}
        extensions["status_code"] = status
        if code is not None:
            extensions["error_code"] = code

        super().__init__(
            message=message,
            nodes=nodes,
            source=source,
            positions=positions,
            path=path,
            original_error=original_error,
            extensions=extensions,
        )


class GraphQLErrorGroup(ExceptionGroup):
    """Exception group for GraphQL errors."""

    msg: ClassVar[str] = ""
    error_formatter = ErrorMessageFormatter()

    # Need to override both __new__ and __init__ so that keyword arguments can be used.

    def __new__(
        cls,
        errors: Sequence[GraphQLError | GraphQLErrorGroup],
        *,
        msg: str = "",
        **kwargs: Any,
    ) -> Self:
        msg = msg or cls.msg
        if kwargs:
            msg = cls.error_formatter.format(msg, **kwargs)
        return super().__new__(GraphQLErrorGroup, msg, errors)  # type: ignore[return-value]

    def __init__(self, errors: Sequence[GraphQLError | GraphQLErrorGroup], *, msg: str = "", **kwargs: Any) -> None:
        super().__init__(msg, errors)

    def __str__(self) -> str:
        return "\n\n".join(str(error) for error in self.flatten())

    def flatten(self) -> Generator[GraphQLError, None, None]:
        """Iterate errors while flattening nested errors groups."""
        for error in self.exceptions:
            if isinstance(error, GraphQLErrorGroup):
                yield from error.flatten()
            elif isinstance(error, GraphQLError):
                yield error

    def located(self, path: list[str | int] | None = None, nodes: list[Node] | None = None) -> Self:
        """Set location information to all errors."""
        for error in self.flatten():
            if not error.path:
                error.path = path
            if not error.nodes:
                error.nodes = nodes
        return self


class GraphQLAsyncAtomicMutationNotSupportedError(GraphQLStatusError):
    """Error raised when a trying to use atomic mutations with async views."""

    msg = "Atomic mutations are not supported when using async views."
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.ASYNC_ATOMIC_MUTATION_NOT_SUPPORTED


class GraphQLAsyncNotSupportedError(GraphQLStatusError):
    """Error raised when a GraphQL request is made asynchronously."""

    msg = "GraphQL execution failed to complete synchronously."
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.ASYNC_NOT_SUPPORTED


class GraphQLDataLoaderDidNotReturnSortedSequenceError(GraphQLStatusError):
    """Error raised when a data loader returns a non-sorted sequence."""

    msg = "DataLoader returned wrong type of object, got '{got:name}' but expected 'list' or 'tuple'"
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.DATA_LOADER_DID_NOT_RETURN_SORTED_SEQUENCE


class GraphQLDataLoaderPrimingError(GraphQLStatusError):  # TODO: Test
    """Error raised when a trying to prime keys and values of different lengths."""

    msg = "Cannot prime DataLoader from {keys} keys to {values} values"
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.DATA_LOADER_PRIMING_ERROR


class GraphQLDataLoaderWrongNumberOfValuesReturnedError(GraphQLStatusError):
    """Error raised when a data loader returns the wrong number of values."""

    msg = "Wrong number of values returned from a DataLoader, got {got} but expected {expected}"
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.DATA_LOADER_WRONG_NUMBER_OF_VALUES_RETURNED


class GraphQLDuplicatePrimaryKeysError(GraphQLStatusError):
    """Error raised when bulk update did not receive primary keys for all input dicts."""

    msg = "Bulk update received instances with duplicate primary keys: {duplicates}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.DUPLICATE_PRIMARY_KEYS


class GraphQLDuplicateTypeError(GraphQLStatusError):
    """Error raised when trying to create a type in the GraphQL schema with the same name as an existing type."""

    msg = (
        "GraphQL schema already has a known type with the name '{name}': '{type_existing!r}'. "
        "Cannot add a new type '{type_new!r}'."
    )
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.DUPLICATE_TYPE


class GraphQLFilePlacingError(GraphQLStatusError):
    """Error raised when placing uploaded files to request data fails."""

    msg = "Value '{value}' in file map does not lead to a null value."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.FILE_NOT_FOUND


class GraphQLFileNotFoundError(GraphQLStatusError):
    """Error raised when a file is not found in the GraphQL request files map."""

    msg = "File for path '{key}' not found in request files."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.FILE_NOT_FOUND


class GraphQLRequestMultipleOperationsNoOperationNameError(GraphQLStatusError):
    """
    Error raised when user tries to execute multiple operations
    through an HTTP GET request without an operation name.
    """

    msg = "Must provide operation name if query contains multiple operations."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_OPERATION_NAME


class GraphQLGetRequestNonQueryOperationError(GraphQLStatusError):
    """Error raised when user tries to execute non-query operations through an HTTP GET request."""

    msg = "Only query operations are allowed on GET requests."
    status = HTTPStatus.METHOD_NOT_ALLOWED
    code = UndineErrorCodes.INVALID_OPERATION_FOR_METHOD


class GraphQLRequestNoOperationError(GraphQLStatusError):
    """Error raised when no operation definition can be found in the request."""

    msg = "Must provide an operation."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NO_OPERATION


class GraphQLRequestOperationNotFoundError(GraphQLStatusError):
    """Error raised when operation matching given operation name cannot be found in the request."""

    msg = "Unknown operation named '{operation_name}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.OPERATION_NOT_FOUND


class GraphQLInvalidInputDataError(GraphQLStatusError):
    """Error raised when a mutation receives invalid input data for some input."""

    msg = "Invalid input data for field {field_name!r}: {data!r}"
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.INVALID_INPUT_DATA


class GraphQLInvalidOrderDataError(GraphQLStatusError):
    """Error raised when a mutation receives invalid order data for an OrderSet."""

    msg = (
        "Order data contains ordering value '{enum_value}' but OrderSet '{orderset:dotpath}' "
        "doesn't have support an order with that name."
    )
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.INVALID_ORDER_DATA


class GraphQLMissingCalculationArgumentError(GraphQLStatusError):
    """Error raised if a calculation argument is missing."""

    msg = "Missing calculation argument '{arg}' for calculation '{name}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_CALCULATION_ARGUMENT


class GraphQLMissingContentTypeError(GraphQLStatusError):
    """Error raised when a request is made without a content type."""

    msg = "Must provide a 'Content-Type' header."
    status = HTTPStatus.UNSUPPORTED_MEDIA_TYPE
    code = UndineErrorCodes.CONTENT_TYPE_MISSING


class GraphQLMissingFileMapError(GraphQLStatusError):
    """Error raised when parsing file upload data doesn't contain a `map` files mapping."""

    msg = "File upload must contain an `map` value."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_FILE_MAP


class GraphQLMissingLookupFieldError(GraphQLStatusError):
    """Error raised when a lookup field is missing from the mutation input data for fetching the mutated instance."""

    msg = (
        "Input data is missing value for the mutation lookup field '{key}'. "
        "Cannot fetch '{model:dotpath}' object for mutation."
    )
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.LOOKUP_VALUE_MISSING


class GraphQLMissingOperationsError(GraphQLStatusError):
    """Error raised when parsing file upload data doesn't contain an `operations` data mapping."""

    msg = "File upload must contain an `operations` value."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_OPERATIONS


class GraphQLMissingDocumentIDError(GraphQLStatusError):
    """Error raised if persisted document id are missing from the request."""

    msg = "Request data must contain a `documentId` string identifying a persisted document."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_GRAPHQL_DOCUMENT_PARAMETER


class GraphQLMissingQueryError(GraphQLStatusError):
    """Error raised when query string is missing from the request."""

    msg = "Request data must contain a `query` string describing the graphql document."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_GRAPHQL_QUERY_PARAMETER


class GraphQLMissingQueryAndDocumentIDError(GraphQLStatusError):
    """Error raised both neither the query string and persisted document id are missing from the request."""

    msg = (
        "Request data must contain either a `query` string describing the graphql document "
        "or a `documentId` string identifying a persisted document."
    )
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_GRAPHQL_QUERY_AND_DOCUMENT_PARAMETERS


class GraphQLMissingInstancesToDeleteError(GraphQLStatusError):
    """Error raised when bulk delete cannot find all instances given to it for deletion."""

    msg = "Expected {given} instances to delete, but found {to_delete}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MISSING_INSTANCES_TO_DELETE


class GraphQLModelConstraintViolationError(GraphQLStatusError):
    """Error raised when a request is made with an unsupported content type."""

    msg = "Model constraint violation"
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MODEL_CONSTRAINT_VIOLATION


class GraphQLModelNotFoundError(GraphQLStatusError):
    """Error raised when a model lookup fails to find a matching row."""

    msg = "Primary key {pk!r} on model '{model:dotpath}' did not match any row."
    status = HTTPStatus.NOT_FOUND
    code = UndineErrorCodes.MODEL_INSTANCE_NOT_FOUND


class GraphQLModelsNotFoundError(GraphQLStatusError):
    """Error raised when model lookup doesn't find expected number of matching rows."""

    msg = "Primary keys {missing:comma_sep_and} on model '{model:dotpath}' did not match any row."
    status = HTTPStatus.NOT_FOUND
    code = UndineErrorCodes.MODEL_INSTANCE_NOT_FOUND


class GraphQLMutationInputNotFoundError(GraphQLStatusError):
    """Error raised when mutation receives input data for a field that doesn't exist in the MutationType."""

    msg = (
        "Input data contains data for field '{field_name}' but MutationType '{mutation_type:dotpath}' "
        "doesn't have an input with that name."
    )
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.INVALID_INPUT_DATA


class GraphQLMutationInstanceLimitError(GraphQLStatusError):
    """Error raised when mutation contains too many objects."""

    msg = "Cannot mutate more than {limit} objects in a single mutation (counted {count})."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MUTATION_TOO_MANY_OBJECTS


class GraphQLMutationTreeModelMismatchError(GraphQLStatusError):
    """Error raised when trying to merge mutation nodes with different models."""

    msg = "Cannot merge MutationNodes for different models: '{model_1:dotpath}' and '{model_2:dotpath}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.MUTATION_TREE_MODEL_MISMATCH


class GraphQLNodeIDFieldTypeError(GraphQLStatusError):
    """Error raised when a Node request `id` field type is not the Global ID type."""

    msg = "The 'id' field of the object type '{typename}' must be of type 'ID' to comply with the 'Node' interface."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_ID_NOT_GLOBAL_ID


class GraphQLNodeInterfaceMissingError(GraphQLStatusError):
    """Error raised when a Node request ObjectType does not implement the Node interface."""

    msg = "Object type '{typename}' must implement the 'Node' interface."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_INTERFACE_MISSING


class GraphQLNodeInvalidGlobalIDError(GraphQLStatusError):
    """Error raised when a Node request `id` is not a valid Global ID."""

    msg = "'{value}' is not a valid Global ID."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_INVALID_GLOBAL_ID


class GraphQLNodeMissingIDFieldError(GraphQLStatusError):
    """Error raised when a Node request ObjectType does not contain an `id` field."""

    msg = "The object type '{typename}' doesn't have an 'id' field."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_QUERY_TYPE_ID_FIELD_MISSING


class GraphQLNodeObjectTypeMissingError(GraphQLStatusError):
    """Error raised when a Node request `id` is for an unrecognized ObjectType."""

    msg = "Object type '{typename}' does not exist in schema."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_MISSING_OBJECT_TYPE


class GraphQLNodeQueryTypeMissingError(GraphQLStatusError):
    """Error raised when a Node request ObjectType does not contain an extension for it's undine QueryType."""

    msg = "Cannot find undine QueryType from object type '{typename}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_QUERY_TYPE_MISSING


class GraphQLNodeTypeNotObjectTypeError(GraphQLStatusError):
    """Error raised when a Node request `id` is for an unrecognized ObjectType."""

    msg = "Node ID type '{typename}' is not an object type."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.NODE_TYPE_NOT_OBJECT_TYPE


class GraphQLNoExecutionResultError(GraphQLStatusError):
    """Error raised when no execution result exists after the GraphQL execution."""

    msg = "No execution result after GraphQL operation."
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.NO_EXECUTION_RESULT


class GraphQLOptimizerError(GraphQLStatusError):
    """Error raised during the optimization compilation process."""

    msg = "GraphQL optimization failed."
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.OPTIMIZER_ERROR


class GraphQLPaginationArgumentValidationError(GraphQLStatusError):
    """Error raised for invalid pagination arguments."""

    msg = "Invalid pagination arguments."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.INVALID_PAGINATION_ARGUMENTS


class GraphQLPermissionError(GraphQLStatusError):
    """Error raised when a permission check fails."""

    msg = "Permission denied."  # Users should add their own message.
    status = HTTPStatus.FORBIDDEN
    code = UndineErrorCodes.PERMISSION_DENIED


class GraphQLPersistedDocumentNotFoundError(GraphQLStatusError):
    """Error raised when a persisted document matching some document id does not exist."""

    msg = "Persisted document {document_id!r} not found."
    status = HTTPStatus.NOT_FOUND
    code = UndineErrorCodes.PERSISTED_DOCUMENT_NOT_FOUND


class GraphQLPersistedDocumentsNotSupportedError(GraphQLStatusError):
    """Error raised when trying to execute a persisted document without the correct app installed."""

    msg = "Server does not support persisted documents."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.PERSISTED_DOCUMENTS_NOT_SUPPORTED


class GraphQLPrimaryKeysMissingError(GraphQLStatusError):
    """Error raised when bulk update did not receive primary keys for all input dicts."""

    msg = "Bulk update missing primary keys for objects: Got {got}, expected {expected}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.PRIMARY_KEYS_MISSING


class GraphQLRelationMultipleInstancesError(GraphQLStatusError):
    """
    Error raised when trying to change the reverse one-to-one relation
    to another instance without changing the current related object.
    Should only happen as a misconfiguration of the related mutation type.
    """

    msg = (
        "Field '{model:dotpath}.{field_name}' is a one-to-one relation, "
        "but trying to set a new instance to it without handling the current instance. "
        "Check if the related mutation action is set to 'ignore', which is not allowed."
    )
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.FIELD_ONE_TO_ONE_CONSTRAINT_VIOLATION


class GraphQLRelationNotNullableError(GraphQLStatusError):
    """Error raised when trying to update a relation to null, but the relation is not nullable."""

    msg = "Field '{model:dotpath}.{field_name}' is not nullable. Existing relation cannot be set to null."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.RELATION_NOT_NULLABLE


class GraphQLFieldNotNullableError(GraphQLStatusError):
    """Error raised when field result is null, but the field is not nullable."""

    msg = "'{typename}.{field_name}' returned null, but field is not nullable."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.FIELD_NOT_NULLABLE


class GraphQLRequestDecodingError(GraphQLStatusError):
    """Error raised when a request content cannot be decoded to python data."""

    msg = "Could not decode request."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.REQUEST_DECODING_ERROR


class GraphQLRequestParseError(GraphQLStatusError):
    """Error raised when a request content cannot be parsed to the expected format."""

    msg = "Could not parse request."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.REQUEST_PARSE_ERROR


class GraphQLScalarConversionError(GraphQLStatusError):
    """Error raised when a value cannot be converted to a GraphQL type."""

    msg = "'{typename}' cannot represent value {value}: {error}"
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.SCALAR_CONVERSION_ERROR


class GraphQLScalarInvalidValueError(GraphQLStatusError):
    """Error raised when a scalar cannot parse or serialize a given value."""

    msg = "Value is not a valid {typename}"
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.SCALAR_INVALID_VALUE


class GraphQLScalarTypeNotSupportedError(GraphQLStatusError):
    """Error raised when a scalar does not support a given type."""

    msg = "Type '{input_type:dotpath}' is not supported"
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.SCALAR_TYPE_NOT_SUPPORTED


class GraphQLSubscriptionNoEventStreamError(GraphQLStatusError):
    """Error raised when a subscription does not return an event stream."""

    msg = "Subscription did not return an event stream"
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.NO_EVENT_STREAM


class GraphQLSubscriptionTimeoutError(GraphQLStatusError):
    """Error raised when a subscription times out."""

    msg = "Subscription timed out"
    status = HTTPStatus.REQUEST_TIMEOUT
    code = UndineErrorCodes.SUBSCRIPTION_TIMEOUT


class GraphQLTooManyFiltersError(GraphQLStatusError):
    """Error raised when too many filters are used for a single `FilterSet`."""

    msg = "{name!r} received {filter_count} filters which is more than the maximum allowed of {max_count}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.TOO_MANY_FILTERS


class GraphQLTooManyOrdersError(GraphQLStatusError):
    """Error raised when too many orders are used for a single `OrderSet`."""

    msg = "{name!r} received {filter_count} orders which is more than the maximum allowed of {max_count}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.TOO_MANY_ORDERS


class GraphQLTypedDictAnnotatedIncorrectMetadataError(GraphQLStatusError):
    """Error raised when a TypedDict key is annotated with an incorrect metadata."""

    msg = "TypedDict '{typed_dict:name}' key '{field}' Annotated metadata must be an instance of '{expected:dotpath}'."
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.TYPED_DICT_ANNOTATED_INCORRECT_METADATA


class GraphQLUnionResolveTypeInvalidValueError(GraphQLStatusError):
    """Error raised when a union resolver fails to resolve a type for an invalid value."""

    msg = "Union '{name}' doesn't support {value!r} of type '{value:type}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.UNION_RESOLVE_TYPE_INVALID_VALUE


class GraphQLUnionResolveTypeModelNotFoundError(GraphQLStatusError):
    """Error raised when a union resolver fails to resolve a type for a given model."""

    msg = "Union '{name}' doesn't contain a type for model '{model:dotpath}'."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.UNION_RESOLVE_TYPE_MODEL_NOT_FOUND


class GraphQLUnexpectedError(GraphQLStatusError):
    """Error raised when an unexpected error occurs."""

    msg = "Unexpected error in GraphQL execution"
    status = HTTPStatus.INTERNAL_SERVER_ERROR
    code = UndineErrorCodes.UNEXPECTED_ERROR


class GraphQLUnexpectedCalculationArgumentError(GraphQLStatusError):
    """Error raised if a calculation argument is unexpected."""

    msg = "Unexpected calculation arguments for field '{name}': {kwargs}."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.UNEXPECTED_CALCULATION_ARGUMENT


class GraphQLUnsupportedContentTypeError(GraphQLStatusError):
    """Error raised when a request is made with an unsupported content type."""

    msg = "'{content_type}' is not a supported content type."
    status = HTTPStatus.UNSUPPORTED_MEDIA_TYPE
    code = UndineErrorCodes.UNSUPPORTED_CONTENT_TYPE


class GraphQLUseWebSocketsForSubscriptionsError(GraphQLStatusError):
    """Error raised when a subscription request is made using HTTP."""

    msg = "Subscriptions do not support HTTP. Please use WebSockets."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.USE_WEBSOCKETS_FOR_SUBSCRIPTIONS


class GraphQLValidationError(GraphQLStatusError):
    """Error meant to be raised for validation errors during mutations."""

    msg = "Validation error."  # Users should add their own message.
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.VALIDATION_ERROR


class GraphQLValidationAbortedError(GraphQLStatusError):
    """Error raised when the validation step of the GraphQL request is aborted, e.g., due to too many errors."""

    msg = "Too many validation errors, error limit reached. Validation aborted."
    status = HTTPStatus.BAD_REQUEST
    code = UndineErrorCodes.VALIDATION_ABORTED


# WebSocket errors


class WebSocketError(Exception):
    """Base class for all websocket errors."""

    reason: str
    code: GraphQLWebSocketCloseCode

    error_formatter = ErrorMessageFormatter()

    def __init__(
        self,
        reason: str | None = None,
        code: GraphQLWebSocketCloseCode | int | None = None,
        **kwargs: Any,
    ) -> None:
        reason = reason or self.reason
        if kwargs:
            reason = self.error_formatter.format(reason, **kwargs)

        self.reason = reason
        self.code = GraphQLWebSocketCloseCode(code or self.code)


class WebSocketConnectionInitAlreadyInProgressError(WebSocketError):
    reason = "Connection initialisation already in progress"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketConnectionInitForbiddenError(WebSocketError):
    reason = "Forbidden"
    code = GraphQLWebSocketCloseCode.FORBIDDEN


class WebSocketConnectionInitTimeoutError(WebSocketError):
    reason = "Connection initialisation timeout"
    code = GraphQLWebSocketCloseCode.CONNECTION_INITIALISATION_TIMEOUT


class WebSocketEmptyMessageError(WebSocketError):
    reason = "Received empty message from client"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInternalServerError(WebSocketError):
    reason = "Internal server error"
    code = GraphQLWebSocketCloseCode.INTERNAL_SERVER_ERROR


class WebSocketInvalidCompleteMessageOperationIdError(WebSocketError):
    reason = "Complete message 'id' must be a string"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidConnectionInitPayloadError(WebSocketError):
    reason = "ConnectionInit 'payload' must be a valid JSON object"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidJSONError(WebSocketError):
    reason = "WebSocket message must be a valid JSON object"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidPingPayloadError(WebSocketError):
    reason = "Ping 'payload' must be a valid JSON object"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidPongPayloadError(WebSocketError):
    reason = "Pong 'payload' must be a valid JSON object"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidSubscribeOperationIdError(WebSocketError):
    reason = "Subscription 'id' must be a string"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketInvalidSubscribePayloadError(WebSocketError):
    reason = "Subscription 'payload' must be a valid JSON object"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketMissingCompleteMessageOperationIdError(WebSocketError):
    reason = "Complete message must contain an 'id' field"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketMissingSubscribeOperationIdError(WebSocketError):
    reason = "Subscribe message must contain an 'id' field"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketMissingSubscribePayloadError(WebSocketError):
    reason = "Subscribe message must contain an 'payload' field"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketTooManyInitialisationRequestsError(WebSocketError):
    reason = "Too many initialisation requests"
    code = GraphQLWebSocketCloseCode.TOO_MANY_INITIALISATION_REQUESTS


class WebSocketSubscriberForOperationIdAlreadyExistsError(WebSocketError):
    reason = "Subscriber for {id} already exists"
    code = GraphQLWebSocketCloseCode.SUBSCRIBER_ALREADY_EXISTS


class WebSocketTypeMissingError(WebSocketError):
    reason = "WebSocket message must contain a 'type' field"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketUnauthorizedError(WebSocketError):
    reason = "Unauthorized"
    code = GraphQLWebSocketCloseCode.UNAUTHORIZED


class WebSocketUnknownMessageTypeError(WebSocketError):
    reason = "Unknown message type: {type!r}"
    code = GraphQLWebSocketCloseCode.BAD_REQUEST


class WebSocketUnsupportedSubProtocolError(WebSocketError):
    reason = "Subprotocol not acceptable"
    code = GraphQLWebSocketCloseCode.SUBPROTOCOL_NOT_ACCEPTABLE


class WebSocketConnectionClosedError(WebSocketError):
    reason = "Connection closed"
    code = GraphQLWebSocketCloseCode.NORMAL_CLOSURE



from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING
from unittest.mock import patch

from django.db.models import ManyToManyField
from django.db.models.fields.related_descriptors import (
    _filter_prefetch_queryset,  # type: ignore[attr-defined] # noqa: PLC2701
)

from undine.settings import undine_settings

if TYPE_CHECKING:
    from django.db.models import ManyToManyRel, Model, QuerySet

    from undine.typing import PrefetchHackCacheType, TModel

__all__ = [
    "evaluate_with_prefetch_hack_async",
    "evaluate_with_prefetch_hack_sync",
    "register_for_prefetch_hack",
]


_PATH = f"{_filter_prefetch_queryset.__module__}.{_filter_prefetch_queryset.__name__}"


def evaluate_with_prefetch_hack_sync(queryset: QuerySet[TModel]) -> list[TModel]:
    """Evaluates the given queryset with the prefetch hack applied."""
    with patch(_PATH, side_effect=_prefetch_hack):
        return list(queryset)  # If the optimizer did its job, the database query is executed here


async def evaluate_with_prefetch_hack_async(queryset: QuerySet[TModel]) -> list[TModel]:
    """Evaluates the given queryset with the prefetch hack applied."""
    with patch(_PATH, side_effect=_prefetch_hack):
        return [inst async for inst in queryset]  # If the optimizer did its job, the database query is executed here


def register_for_prefetch_hack(queryset: QuerySet, field: ManyToManyField | ManyToManyRel) -> None:
    """
    Registers the through table of a many-to-many field for the prefetch hack.
    See `_prefetch_hack` for more information.
    """
    related_model: type[Model] = field.related_model  # type: ignore[assignment]
    db_table = related_model._meta.db_table
    field_name = field.remote_field.name

    forward_field: ManyToManyField
    forward_field = field if isinstance(field, ManyToManyField) else field.remote_field  # type: ignore[assignment]
    through = forward_field.m2m_db_table()

    key = undine_settings.PREFETCH_HACK_CACHE_KEY
    cache: PrefetchHackCacheType = queryset._hints.setdefault(key, defaultdict(lambda: defaultdict(set)))  # type: ignore[attr-defined]
    cache[db_table][field_name].add(through)


def _prefetch_hack(queryset: QuerySet, field_name: str, instances: list[Model]) -> QuerySet:
    """
    Patches the prefetch mechanism to not create duplicate joins in the SQL query.
    This is needed due to how filtering with many-to-many relations is implemented in Django,
    which creates new joins for consecutive filters for the same relation.
    See: https://docs.djangoproject.com/en/stable/topics/db/queries/#spanning-multi-valued-relationships

    For nested connection fields, a window function with a partition on the many-to-many field
    is used to limit the queryset. This adds an OUTER join for the through table of the many-to-many field.
    Then, when the prefetch queryset is filtered by `_filter_prefetch_queryset` to just the instances
    from the parent model, an INNER join is added to the through table. This creates unnecessary duplicates
    in the SQL query, which messes up the window function's partitioning. Therefore, this hack is needed
    to prevent the INNER join from being added.
    """
    key = undine_settings.PREFETCH_HACK_CACHE_KEY
    cache: PrefetchHackCacheType | None = queryset._hints.pop(key, None)  # type: ignore[attr-defined]
    if cache is not None:
        #
        # `filter_is_sticky` is set here just to prevent the `used_aliases` from being cleared
        # when the queryset is cloned for filtering in `_filter_prefetch_queryset`.
        # See: `django.db.models.sql.query.Query.chain`.
        queryset.query.filter_is_sticky = True
        #
        # Add the registered through tables for a given model and field to the Query's `used_aliases`.
        # This is passed along during the filtering that happens as a part of `_filter_prefetch_queryset`,
        # until `django.db.models.sql.query.Query.join`, which has access to it with its `reuse` argument.
        # There, this should prevent the method from adding a duplicate join.
        queryset.query.used_aliases = cache[queryset.model._meta.db_table][field_name]

    return _filter_prefetch_queryset(queryset, field_name, instances)



from __future__ import annotations

from django.core.validators import validate_email

from ._definition import ScalarType

__all__ = [
    "GraphQLEmail",
    "email_scalar",
]


email_scalar: ScalarType[str, str] = ScalarType(
    name="Email",
    description="Represents a valid email address.",
    specified_by_url="https://datatracker.ietf.org/doc/html/rfc5322#section-3.4.1",
)

GraphQLEmail = email_scalar.as_graphql_scalar()


@email_scalar.serialize.register
@email_scalar.parse.register
def _(value: str) -> str:
    if value:
        validate_email(value)
    return value



from __future__ import annotations

import datetime
import decimal
import uuid
from enum import Enum
from types import FunctionType
from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey, GenericRel, GenericRelation
from django.db.models import F, Field, TextChoices
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)
from django.db.models.query_utils import DeferredAttribute, Q
from graphql import GraphQLNonNull, GraphQLType

from undine import Calculation, InterfaceField, QueryType
from undine import Field as UndineField
from undine.converters import convert_to_field_ref, convert_to_graphql_argument_map, convert_to_graphql_type, is_many
from undine.dataclasses import LazyGenericForeignKey, LazyLambda, LazyRelation, TypeRef
from undine.exceptions import (
    InterfaceFieldDoesNotExistError,
    InterfaceFieldTypeMismatchError,
    ModelFieldDoesNotExistError,
)
from undine.optimizer.optimizer import OptimizationData
from undine.pagination import OffsetPagination
from undine.relay import Connection, Node
from undine.settings import undine_settings
from undine.typing import CombinableExpression, GQLInfo, Lambda, ToManyField, ToOneField
from undine.utils.graphql.utils import get_arguments, get_queried_field_name
from undine.utils.model_utils import determine_output_field, get_model_field


@convert_to_field_ref.register
def _(_: None, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return convert_to_field_ref(field, **kwargs)


@convert_to_field_ref.register
def _(_: str, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return convert_to_field_ref(field, **kwargs)


@convert_to_field_ref.register
def _(ref: FunctionType, **kwargs: Any) -> Any:
    return ref


@convert_to_field_ref.register
def _(ref: Lambda, **kwargs: Any) -> Any:
    return LazyLambda(callback=ref)


@convert_to_field_ref.register
def _(ref: CombinableExpression, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    ref.output_field = determine_output_field(ref, model=caller.query_type.__model__)

    user_func = caller.optimizer_func

    def optimizer_func(field: UndineField, data: OptimizationData, info: GQLInfo) -> None:
        if user_func is not None:
            user_func(field, data, info)

        data.annotations[field.name] = ref

    caller.optimizer_func = optimizer_func
    return ref


@convert_to_field_ref.register
def _(ref: Q, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]

    user_func = caller.optimizer_func

    def optimizer_func(field: UndineField, data: OptimizationData, info: GQLInfo) -> None:
        if user_func is not None:
            user_func(field, data, info)

        data.annotations[field.name] = ref

    caller.optimizer_func = optimizer_func
    return ref


@convert_to_field_ref.register
def _(ref: F, **kwargs: Any) -> Any:
    return convert_to_field_ref(ref.name, **kwargs)


@convert_to_field_ref.register
def _(ref: Field, **kwargs: Any) -> Any:
    return ref


@convert_to_field_ref.register
def _(ref: ToOneField | ToManyField, **kwargs: Any) -> Any:
    return LazyRelation(field=ref)


@convert_to_field_ref.register
def _(ref: DeferredAttribute | ForwardManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_field_ref(ref.field, **kwargs)


@convert_to_field_ref.register
def _(ref: ReverseManyToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_field_ref(ref.rel, **kwargs)


@convert_to_field_ref.register
def _(ref: ReverseOneToOneDescriptor, **kwargs: Any) -> Any:
    return convert_to_field_ref(ref.related, **kwargs)


@convert_to_field_ref.register
def _(ref: ManyToManyDescriptor, **kwargs: Any) -> Any:
    return convert_to_field_ref(ref.rel if ref.reverse else ref.field, **kwargs)


@convert_to_field_ref.register
def _(ref: GraphQLType, **kwargs: Any) -> Any:
    return ref


@convert_to_field_ref.register
def _(ref: type[str], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[bool], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[int], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[float], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[decimal.Decimal], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[datetime.datetime], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[datetime.date], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[datetime.time], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[datetime.timedelta], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[uuid.UUID], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[Enum | TextChoices], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[list], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[dict], **kwargs: Any) -> Any:
    return TypeRef(value=ref)


@convert_to_field_ref.register
def _(ref: type[Calculation], **kwargs: Any) -> Any:
    return ref


@convert_to_field_ref.register
def _(ref: type[QueryType], **kwargs: Any) -> Any:
    return ref


@convert_to_field_ref.register
def _(ref: GenericRelation, **kwargs: Any) -> Any:
    return LazyRelation(field=ref)


@convert_to_field_ref.register
def _(ref: GenericRel, **kwargs: Any) -> Any:
    return LazyRelation(field=ref.field)


@convert_to_field_ref.register
def _(ref: GenericForeignKey, **kwargs: Any) -> Any:
    return LazyGenericForeignKey(field=ref)


@convert_to_field_ref.register
def _(ref: Connection, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    caller.extensions[undine_settings.CONNECTION_EXTENSIONS_KEY] = ref
    return ref


@convert_to_field_ref.register
def _(ref: OffsetPagination, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    caller.extensions[undine_settings.OFFSET_PAGINATION_EXTENSIONS_KEY] = ref
    return ref


@convert_to_field_ref.register
def _(ref: InterfaceField, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]

    try:
        field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    except ModelFieldDoesNotExistError as error:
        raise InterfaceFieldDoesNotExistError(
            field=caller.schema_name,
            interface=ref.interface_type,
            model=caller.query_type.__model__,
        ) from error

    field_type = convert_to_graphql_type(field, model=caller.query_type.__model__)
    if not field.null:
        field_type = GraphQLNonNull(field_type)

    many = is_many(field, model=caller.query_type.__model__, name=caller.field_name)
    args = convert_to_graphql_argument_map(field, many=many)

    if ref.output_type == field_type and ref.args == args:
        return ref

    # Node interface is special, since it converts primary key into string ID.
    # In this case the types don't match, but the field will work.
    if ref.interface_type is Node:
        return ref

    # The Node interface id field might also be inherited by another interface.
    if (
        Node in ref.interface_type.__interfaces__
        and caller.schema_name == "id"
        and ref.output_type == Node.id.output_type
    ):
        return ref

    raise InterfaceFieldTypeMismatchError(
        field=caller.schema_name,
        interface=ref.interface_type,
        output_type=ref.output_type,
        field_type=field_type,
    )


@convert_to_field_ref.register
def _(ref: type[Calculation], **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]

    user_func = caller.optimizer_func

    def optimizer_func(field: UndineField, data: OptimizationData, info: GQLInfo) -> None:
        if user_func is not None:
            user_func(field, data, info)

        arg_values = get_arguments(info)
        field_name = get_queried_field_name(field.name, info)
        calculation = ref(field_name, **arg_values)
        data.field_calculations.append(calculation)

    caller.optimizer_func = optimizer_func
    return ref



from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING

from undine.converters import convert_model_field_to_python_type
from undine.dataclasses import RelInfo
from undine.typing import RelationType
from undine.utils.model_utils import generic_foreign_key_for_generic_relation, get_related_name

if TYPE_CHECKING:
    from django.contrib.contenttypes.fields import GenericRelation
    from django.db.models import Field, ForeignKey, Model

    from undine.typing import ForwardField, GenericField, RelatedField, ReverseField

__all__ = [
    "parse_model_relation_info",
]


@cache
def parse_model_relation_info(*, model: type[Model]) -> dict[str, RelInfo]:
    relation_info: dict[str, RelInfo] = {}

    for field in model._meta.get_fields():
        if not field.is_relation:
            continue

        related_field: RelatedField | GenericField = field  # type: ignore[assignment]
        relation_type = RelationType.for_related_field(related_field)

        if relation_type.is_generic_relation:
            generic_rel_field: GenericRelation = related_field  # type: ignore[assignment]
            generic_fk_field = generic_foreign_key_for_generic_relation(generic_rel_field)
            name = generic_rel_field.name

            fk_field: Field = generic_fk_field.model._meta.get_field(generic_fk_field.fk_field)  # type: ignore[assignment]
            ct_field: ForeignKey = generic_fk_field.model._meta.get_field(generic_fk_field.ct_field)  # type: ignore[assignment]

            relation_info[name] = RelInfo(
                relation_type=relation_type,
                #
                # Source details
                field_name=name,
                model=model,
                model_pk_type=convert_model_field_to_python_type(fk_field),  # Type of the fk field
                nullable=True,  # Reverse fields are always nullable.
                #
                # Target details
                related_name=generic_fk_field.name,
                related_model=generic_fk_field.model,
                related_model_pk_type=convert_model_field_to_python_type(generic_fk_field.model._meta.pk),
                related_nullable=fk_field.null and ct_field.null,
            )
            continue

        if relation_type.is_generic_foreign_key:
            # For GenericForeignKey, there are multiple related models,
            # so we don't have a single model or related name.
            generic_fk_field = related_field  # type: ignore[assignment]
            name = generic_fk_field.name

            fk_field = generic_fk_field.model._meta.get_field(generic_fk_field.fk_field)  # type: ignore[assignment]
            ct_field = generic_fk_field.model._meta.get_field(generic_fk_field.ct_field)  # type: ignore[assignment]

            relation_info[name] = RelInfo(
                relation_type=relation_type,
                #
                # Source details
                field_name=name,
                model=model,
                model_pk_type=convert_model_field_to_python_type(model._meta.pk),
                nullable=fk_field.null and ct_field.null,
                #
                # Target details
                #
                # For GenericForeignKey, there are multiple related models,
                # so we don't have a single model or related name.
                related_name=None,
                related_model=None,
                related_model_pk_type=convert_model_field_to_python_type(
                    fk_field
                ),  # Type of the fk field, not related model pk.
                related_nullable=True,  # Reverse fields are always nullable.
            )
            continue

        if relation_type.is_forward:
            forward_field: ForwardField = related_field  # type: ignore[assignment]
            related_model = forward_field.remote_field.model
            name = forward_field.name

            nullable = True if relation_type.is_many_to_many else forward_field.null

            is_self_relation = model == related_model

            relation_info[name] = RelInfo(
                relation_type=relation_type,
                #
                # Source details
                field_name=name,
                model=model,
                model_pk_type=convert_model_field_to_python_type(model._meta.pk),
                nullable=nullable,
                #
                # Target details
                related_name=get_related_name(field if is_self_relation else field.remote_field),  # type: ignore[arg-type]
                related_model=related_model,
                related_model_pk_type=convert_model_field_to_python_type(related_model._meta.pk),
                related_nullable=True,  # Reverse fields are always nullable
            )
            continue

        if relation_type.is_reverse:
            reverse_field: ReverseField = related_field  # type: ignore[assignment]
            forward_field = reverse_field.remote_field  # type: ignore[assignment]
            name = get_related_name(reverse_field)

            nullable = True if relation_type.is_many_to_many else forward_field.null

            relation_info[name] = RelInfo(
                relation_type=relation_type,
                #
                # Source details
                field_name=name,
                model=model,
                model_pk_type=convert_model_field_to_python_type(model._meta.pk),
                nullable=True,  # Reverse fields are always nullable
                #
                # Target details
                related_name=forward_field.name,
                related_model=forward_field.model,
                related_model_pk_type=convert_model_field_to_python_type(forward_field.model._meta.pk),
                related_nullable=nullable,
            )
            continue

        msg = f"Unhandled relation type: {relation_type}"  # pragma: no cover
        raise NotImplementedError(msg)  # pragma: no cover

    return relation_info



from __future__ import annotations

from types import FunctionType, NoneType, UnionType
from typing import Any, Union, get_origin

from graphql import GraphQLNonNull, GraphQLType

from undine.typing import ParametrizedType
from undine.utils.reflection import get_flattened_generic_params, is_not_required_type, is_required_type

from .parse_annotations import parse_first_param_type, parse_return_annotation

__all__ = [
    "parse_is_nullable",
]


def parse_is_nullable(ref: Any, *, is_input: bool = False, total: bool = True) -> bool:
    """
    Determine whether the given reference indicates a nullable GraphQL type or not.

    :param ref: The reference to check.
    :param is_input: Whether the reference is for an input or output type.
    :param total: If the reference is in a TypedDict, whether the TypedDict has totality of not.
    """
    # GraphQL doesn't differentiate between required and non-null...
    if not total:
        if not is_required_type(ref):
            return True

        args = get_flattened_generic_params(ref)
        return NoneType in args

    if is_not_required_type(ref):
        return True

    if isinstance(ref, GraphQLNonNull):
        return False

    if isinstance(ref, GraphQLType):
        return True

    if isinstance(ref, FunctionType):
        ref = parse_first_param_type(ref) if is_input else parse_return_annotation(ref)

    origin = get_origin(ref)
    if origin not in {UnionType, Union, ParametrizedType}:
        return False

    args = get_flattened_generic_params(ref)
    return NoneType in args



from __future__ import annotations

from typing import Any

from django.contrib.contenttypes.fields import GenericForeignKey

from undine import Field as UndineField
from undine import InterfaceField, QueryType
from undine.converters import convert_to_field_complexity
from undine.dataclasses import LazyGenericForeignKey, LazyRelation
from undine.typing import ModelField, ToManyField, ToOneField
from undine.utils.model_utils import get_model_field


@convert_to_field_complexity.register
def _(_: Any, **kwargs: Any) -> int:
    return 0


@convert_to_field_complexity.register
def _(_: ModelField, **kwargs: Any) -> Any:
    return 0


@convert_to_field_complexity.register
def _(_: ToOneField | ToManyField, **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register  # Required for Django<5.1
def _(_: GenericForeignKey, **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register
def _(_: type[QueryType], **kwargs: Any) -> Any:
    return 1


@convert_to_field_complexity.register
def _(ref: LazyRelation, **kwargs: Any) -> Any:
    return convert_to_field_complexity(ref.field, **kwargs)


@convert_to_field_complexity.register
def _(ref: LazyGenericForeignKey, **kwargs: Any) -> Any:
    return convert_to_field_complexity(ref.field, **kwargs)


@convert_to_field_complexity.register
def _(_: InterfaceField, **kwargs: Any) -> Any:
    caller: UndineField = kwargs["caller"]
    field = get_model_field(model=caller.query_type.__model__, lookup=caller.field_name)
    return convert_to_field_complexity(field, **kwargs)



from __future__ import annotations

from django.apps import AppConfig


class UndinePersistedDocumentsConfig(AppConfig):
    """
    Add persisted documents to Undine as defined by the GraphQL over HTTP spec.
    See: https://github.com/graphql/graphql-over-http/blob/persisted-documents-get-url/spec/Appendix%20A%20--%20Persisted%20Documents.md
    """

    name = "undine.persisted_documents"
    label = "persisted_documents"
    verbose_name = "persisted documents"
    default_auto_field = "django.db.models.BigAutoField"



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphql import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLString,
    GraphQLUnionType,
    TypeKind,
    TypeMetaFieldDef,
    introspection_types,
)
from graphql.pyutils import inspect

from undine import InterfaceField

from .undine_extensions import (
    get_undine_calculation_argument,
    get_undine_connection,
    get_undine_directive,
    get_undine_directive_argument,
    get_undine_entrypoint,
    get_undine_field,
    get_undine_filter,
    get_undine_filterset,
    get_undine_input,
    get_undine_interface_field,
    get_undine_interface_type,
    get_undine_mutation_type,
    get_undine_order,
    get_undine_orderset,
    get_undine_query_type,
    get_undine_union_type,
)
from .utils import get_underlying_type
from .validation_rules.one_of_input_object import is_one_of_input_object

if TYPE_CHECKING:
    from collections.abc import Iterable

    from graphql import DirectiveLocation, GraphQLNamedType, GraphQLOutputType, GraphQLSchema, GraphQLType

    from undine import GQLInfo
    from undine.typing import HasGraphQLExtensions


__all__ = [
    "patch_introspection_schema",
]


schema_introspection_type: GraphQLObjectType = introspection_types["__Schema"]  # type: ignore[assignment]
directive_introspection_type: GraphQLObjectType = introspection_types["__Directive"]  # type: ignore[assignment]
directive_location_introspection_type: GraphQLEnumType = introspection_types["__DirectiveLocation"]  # type: ignore[assignment]
type_introspection_type: GraphQLObjectType = introspection_types["__Type"]  # type: ignore[assignment]
field_introspection_type: GraphQLObjectType = introspection_types["__Field"]  # type: ignore[assignment]
input_value_introspection_type: GraphQLObjectType = introspection_types["__InputValue"]  # type: ignore[assignment]
enum_value_introspection_type: GraphQLObjectType = introspection_types["__EnumValue"]  # type: ignore[assignment]
type_kind_introspection_type: GraphQLEnumType = introspection_types["__TypeKind"]  # type: ignore[assignment]


def is_visible(obj: HasGraphQLExtensions, info: GQLInfo) -> bool:  # noqa: PLR0911, PLR0912, PLR0914, PLR0915, C901
    match obj:
        case GraphQLObjectType():
            query_type = get_undine_query_type(obj)
            if query_type is not None:
                return query_type.__is_visible__(info.context)

            connection = get_undine_connection(obj)
            if connection is not None:
                if connection.query_type is not None:
                    return connection.query_type.__is_visible__(info.context)
                if connection.union_type is not None:
                    return connection.union_type.__is_visible__(info.context)
                if connection.interface_type is not None:
                    return connection.interface_type.__is_visible__(info.context)
                return True  # Should never happen

        case GraphQLInputObjectType():
            mutation_type = get_undine_mutation_type(obj)
            if mutation_type is not None:
                return mutation_type.__is_visible__(info.context)

            filterset = get_undine_filterset(obj)
            if filterset is not None:
                return filterset.__is_visible__(info.context)

        case GraphQLInterfaceType():
            interface_type = get_undine_interface_type(obj)
            if interface_type is not None:
                return interface_type.__is_visible__(info.context)

        case GraphQLUnionType():
            union_type = get_undine_union_type(obj)
            if union_type is not None:
                return union_type.__is_visible__(info.context)

        case GraphQLEnumType():
            orderset = get_undine_orderset(obj)
            if orderset is not None:
                return orderset.__is_visible__(info.context)

        case GraphQLDirective():
            directive = get_undine_directive(obj)
            if directive is not None:
                return directive.__is_visible__(info.context)

        case GraphQLField():
            entrypoint = get_undine_entrypoint(obj)
            if entrypoint is not None:
                if entrypoint.visible_func is not None:
                    return entrypoint.visible_func(entrypoint, info.context)
                return True

            field = get_undine_field(obj)
            if field is not None:
                if field.visible_func is not None:
                    return field.visible_func(field, info.context)
                if isinstance(field.ref, InterfaceField):
                    if not field.ref.interface_type.__is_visible__(info.context):
                        return False
                    if field.ref.visible_func is not None:
                        return field.ref.visible_func(field.ref, info.context)
                return True

            interface_field = get_undine_interface_field(obj)
            if interface_field is not None:
                if interface_field.visible_func is not None:
                    return interface_field.visible_func(interface_field, info.context)
                return True

            field_type = get_underlying_type(obj.type)
            return is_visible(field_type, info)

        case GraphQLInputField():
            inpt = get_undine_input(obj)
            if inpt is not None:
                if inpt.visible_func is not None:
                    return inpt.visible_func(inpt, info.context)
                return True

            ftr = get_undine_filter(obj)
            if ftr is not None:
                if ftr.visible_func is not None:
                    return ftr.visible_func(ftr, info.context)
                return True

            input_field_type = get_underlying_type(obj.type)
            return is_visible(input_field_type, info)

        case GraphQLArgument():
            directive_arg = get_undine_directive_argument(obj)
            if directive_arg is not None:
                if directive_arg.visible_func is not None:
                    return directive_arg.visible_func(directive_arg, info.context)
                return True

            calculation_arg = get_undine_calculation_argument(obj)
            if calculation_arg is not None:
                if calculation_arg.visible_func is not None:
                    return calculation_arg.visible_func(calculation_arg, info.context)
                return True

            arg_type = get_underlying_type(obj.type)
            return is_visible(arg_type, info)

        case GraphQLEnumValue():
            order = get_undine_order(obj)
            if order is not None:
                if order.visible_func is not None:
                    return order.visible_func(order, info.context)
                return True

    return True


def patch_introspection_schema() -> None:
    TypeMetaFieldDef.resolve = resolve_type_meta_field_def

    schema_introspection_type._fields = get_schema_fields
    directive_introspection_type._fields = get_directive_fields
    type_introspection_type._fields = get_type_fields
    field_introspection_type._fields = get_field_fields

    # Force re-evaluation of cached properties
    if "fields" in schema_introspection_type.__dict__:
        del schema_introspection_type.__dict__["fields"]

    if "fields" in directive_introspection_type.__dict__:
        del directive_introspection_type.__dict__["fields"]

    if "fields" in type_introspection_type.__dict__:
        del type_introspection_type.__dict__["fields"]

    if "fields" in field_introspection_type.__dict__:
        del field_introspection_type.__dict__["fields"]


def resolve_type_meta_field_def(root: Any, info: GQLInfo, *, name: str) -> GraphQLNamedType | None:
    gql_type: GraphQLNamedType | None = info.schema.get_type(name)
    if gql_type is None:
        return None

    if not is_visible(gql_type, info):
        return None

    return gql_type


# Schema


def get_schema_fields() -> dict[str, GraphQLField]:
    return {
        "description": GraphQLField(
            GraphQLString,
            resolve=resolve_schema_description,
        ),
        "types": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(type_introspection_type))),
            resolve=resolve_schema_types,
            description="A list of all types supported by this server.",
        ),
        "queryType": GraphQLField(
            GraphQLNonNull(type_introspection_type),
            resolve=resolve_schema_query_type,
            description="The type that query operations will be rooted at.",
        ),
        "mutationType": GraphQLField(
            type_introspection_type,
            resolve=resolve_schema_mutation_type,
            description="If this server supports mutation, the type that mutation operations will be rooted at.",
        ),
        "subscriptionType": GraphQLField(
            type_introspection_type,
            resolve=resolve_schema_subscription_type,
            description="If this server support subscription, the type that subscription operations will be rooted at.",
        ),
        "directives": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(directive_introspection_type))),
            resolve=resolve_schema_directives,
            description="A list of all directives supported by this server.",
        ),
    }


def resolve_schema_description(root: GraphQLSchema, info: GQLInfo) -> str | None:
    return root.description


def resolve_schema_types(root: GraphQLSchema, info: GQLInfo) -> Iterable[GraphQLNamedType]:
    return [gql_type for gql_type in root.type_map.values() if is_visible(gql_type, info)]


def resolve_schema_query_type(root: GraphQLSchema, info: GQLInfo) -> GraphQLObjectType:
    return root.query_type  # type: ignore[return-value]


def resolve_schema_mutation_type(root: GraphQLSchema, info: GQLInfo) -> GraphQLObjectType | None:
    return root.mutation_type


def resolve_schema_subscription_type(root: GraphQLSchema, info: GQLInfo) -> GraphQLObjectType | None:
    return root.subscription_type


def resolve_schema_directives(root: GraphQLSchema, info: GQLInfo) -> Iterable[GraphQLDirective]:
    return [directive for directive in root.directives if is_visible(directive, info)]


# Directive


def get_directive_fields() -> dict[str, GraphQLField]:
    return {
        "name": GraphQLField(
            GraphQLNonNull(GraphQLString),
            resolve=resolve_directive_name,
        ),
        "description": GraphQLField(
            GraphQLString,
            resolve=resolve_directive_description,
        ),
        "isRepeatable": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            resolve=resolve_directive_is_repeatable,
        ),
        "locations": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(directive_location_introspection_type))),
            resolve=resolve_directive_locations,
        ),
        "args": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(input_value_introspection_type))),
            args={
                "includeDeprecated": GraphQLArgument(GraphQLBoolean, default_value=False),
            },
            resolve=resolve_directive_args,
        ),
    }


def resolve_directive_name(root: GraphQLDirective, info: GQLInfo) -> str:
    return root.name


def resolve_directive_description(root: GraphQLDirective, info: GQLInfo) -> str | None:
    return root.description


def resolve_directive_is_repeatable(root: GraphQLDirective, info: GQLInfo) -> bool:
    return root.is_repeatable


def resolve_directive_locations(root: GraphQLDirective, info: GQLInfo) -> Iterable[DirectiveLocation]:
    return root.locations


def resolve_directive_args(root: GraphQLDirective, info: GQLInfo, **kwargs: Any) -> list[tuple[str, GraphQLArgument]]:
    args = ((key, arg) for key, arg in root.args.items() if is_visible(arg, info))

    if kwargs["includeDeprecated"]:
        return list(args)

    return [(key, value) for key, value in args if value.deprecation_reason is None]


# Type


def get_type_fields() -> dict[str, GraphQLField]:
    return {
        "kind": GraphQLField(
            GraphQLNonNull(type_kind_introspection_type),
            resolve=resolve_type_kind,
        ),
        "name": GraphQLField(
            GraphQLString,
            resolve=resolve_type_name,
        ),
        "description": GraphQLField(
            GraphQLString,
            resolve=resolve_type_description,
        ),
        "specifiedByURL": GraphQLField(
            GraphQLString,
            resolve=resolve_type_specified_by_url,
        ),
        "fields": GraphQLField(
            GraphQLList(GraphQLNonNull(field_introspection_type)),
            args={
                "includeDeprecated": GraphQLArgument(GraphQLBoolean, default_value=False),
            },
            resolve=resolve_type_fields,
        ),
        "interfaces": GraphQLField(
            GraphQLList(GraphQLNonNull(type_introspection_type)),
            resolve=resolve_type_interfaces,
        ),
        "possibleTypes": GraphQLField(
            GraphQLList(GraphQLNonNull(type_introspection_type)),
            resolve=resolve_type_possible_types,
        ),
        "enumValues": GraphQLField(
            GraphQLList(GraphQLNonNull(enum_value_introspection_type)),
            args={
                "includeDeprecated": GraphQLArgument(GraphQLBoolean, default_value=False),
            },
            resolve=resolve_type_enum_values,
        ),
        "inputFields": GraphQLField(
            GraphQLList(GraphQLNonNull(input_value_introspection_type)),
            args={
                "includeDeprecated": GraphQLArgument(GraphQLBoolean, default_value=False),
            },
            resolve=resolve_type_input_fields,
        ),
        "ofType": GraphQLField(
            type_introspection_type,
            resolve=resolve_type_of_type,
        ),
        "isOneOf": GraphQLField(
            GraphQLBoolean,
            resolve=resolve_type_is_one_of,
        ),
    }


def resolve_type_kind(gql_type: GraphQLType, info: GQLInfo) -> TypeKind:
    match gql_type:
        case GraphQLScalarType():
            return TypeKind.SCALAR
        case GraphQLObjectType():
            return TypeKind.OBJECT
        case GraphQLInterfaceType():
            return TypeKind.INTERFACE
        case GraphQLUnionType():
            return TypeKind.UNION
        case GraphQLEnumType():
            return TypeKind.ENUM
        case GraphQLInputObjectType():
            return TypeKind.INPUT_OBJECT
        case GraphQLList():
            return TypeKind.LIST
        case GraphQLNonNull():
            return TypeKind.NON_NULL

    msg = f"Unexpected type: {inspect(gql_type)}."  # pragma: no cover
    raise TypeError(msg)  # pragma: no cover


def resolve_type_name(gql_type: GraphQLType, info: GQLInfo) -> str | None:
    return getattr(gql_type, "name", None)


def resolve_type_description(gql_type: GraphQLType, info: GQLInfo) -> str | None:
    return getattr(gql_type, "description", None)


def resolve_type_specified_by_url(gql_type: GraphQLType, info: GQLInfo) -> str | None:
    return getattr(gql_type, "specified_by_url", None)


def resolve_type_fields(gql_type: GraphQLType, info: GQLInfo, **kwargs: Any) -> list[tuple[str, GraphQLField]] | None:
    if isinstance(gql_type, (GraphQLObjectType, GraphQLInterfaceType)):
        fields = (
            (key, field)
            for key, field in gql_type.fields.items()
            if is_visible(field, info) and is_visible(get_underlying_type(field.type), info)
        )

        if kwargs["includeDeprecated"]:
            return list(fields)

        return [(key, value) for key, value in fields if value.deprecation_reason is None]

    return None


def resolve_type_interfaces(gql_type: GraphQLType, info: GQLInfo) -> Iterable[GraphQLInterfaceType] | None:
    if isinstance(gql_type, (GraphQLObjectType, GraphQLInterfaceType)):
        return [interface for interface in gql_type.interfaces if is_visible(interface, info)]

    return None


def resolve_type_possible_types(gql_type: GraphQLType, info: GQLInfo) -> Iterable[GraphQLObjectType] | None:
    if isinstance(gql_type, (GraphQLInterfaceType, GraphQLUnionType)):
        object_types = info.schema.get_possible_types(gql_type)
        return [object_type for object_type in object_types if is_visible(object_type, info)]

    return None


def resolve_type_enum_values(
    gql_type: GraphQLType,
    info: GQLInfo,
    **kwargs: Any,
) -> list[tuple[str, GraphQLEnumValue]] | None:
    if isinstance(gql_type, GraphQLEnumType):
        values = ((key, field) for key, field in gql_type.values.items() if is_visible(field, info))

        if kwargs["includeDeprecated"]:
            return list(values)

        return [(key, value) for key, value in values if value.deprecation_reason is None]

    return None


def resolve_type_input_fields(
    gql_type: GraphQLType,
    info: GQLInfo,
    **kwargs: Any,
) -> list[tuple[str, GraphQLInputField]] | None:
    if isinstance(gql_type, GraphQLInputObjectType):
        fields = (
            (key, field)
            for key, field in gql_type.fields.items()
            if is_visible(field, info) and is_visible(get_underlying_type(field.type), info)
        )

        if kwargs["includeDeprecated"]:
            return list(fields)

        return [(key, value) for key, value in fields if value.deprecation_reason is None]

    return None


def resolve_type_of_type(gql_type: GraphQLType, info: GQLInfo) -> GraphQLType | None:
    return getattr(gql_type, "of_type", None)


def resolve_type_is_one_of(gql_type: GraphQLType, info: GQLInfo) -> bool | None:
    if isinstance(gql_type, GraphQLInputObjectType):
        return is_one_of_input_object(gql_type)
    return None


# Field


def get_field_fields() -> dict[str, GraphQLField]:
    return {
        "name": GraphQLField(
            GraphQLNonNull(GraphQLString),
            resolve=resolve_field_name,
        ),
        "description": GraphQLField(
            GraphQLString,
            resolve=resolve_field_description,
        ),
        "args": GraphQLField(
            GraphQLNonNull(GraphQLList(GraphQLNonNull(input_value_introspection_type))),
            args={
                "includeDeprecated": GraphQLArgument(GraphQLBoolean, default_value=False),
            },
            resolve=resolve_field_args,
        ),
        "type": GraphQLField(
            GraphQLNonNull(type_introspection_type),
            resolve=resolve_field_type,
        ),
        "isDeprecated": GraphQLField(
            GraphQLNonNull(GraphQLBoolean),
            resolve=resolve_field_is_deprecated,
        ),
        "deprecationReason": GraphQLField(
            GraphQLString,
            resolve=resolve_field_deprecation_reason,
        ),
    }


def resolve_field_name(item: tuple[str, GraphQLField], info: GQLInfo) -> str:
    return item[0]


def resolve_field_description(item: tuple[str, GraphQLField], info: GQLInfo) -> str | None:
    return item[1].description


def resolve_field_args(
    item: tuple[str, GraphQLField],
    info: GQLInfo,
    **kwargs: Any,
) -> list[tuple[str, GraphQLArgument]]:
    args = ((key, arg) for key, arg in item[1].args.items() if is_visible(arg, info))

    if kwargs["includeDeprecated"]:
        return list(args)

    return [item for item in args if item[1].deprecation_reason is None]


def resolve_field_type(item: tuple[str, GraphQLField], info: GQLInfo) -> GraphQLOutputType:
    return item[1].type


def resolve_field_is_deprecated(item: tuple[str, GraphQLField], info: GQLInfo) -> bool:
    return item[1].deprecation_reason is not None


def resolve_field_deprecation_reason(item: tuple[str, GraphQLField], info: GQLInfo) -> str | None:
    return item[1].deprecation_reason



from __future__ import annotations

from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator, get_available_image_extensions

from undine.utils.text import comma_sep_str

if TYPE_CHECKING:
    from django.core.files.uploadedfile import UploadedFile


__all__ = [
    "validate_file_upload",
    "validate_file_url",
    "validate_image_url",
    "validate_is_allowed_image_extension",
    "validate_url",
]


url_validator = URLValidator()


def validate_url(url: str) -> str:
    url_validator(url)
    return url


def validate_file_url(url: str) -> str:
    url_validator(url)

    extension = Path(url).suffix[1:]
    if not extension:
        msg = "File URLs must have a file extension."
        raise ValidationError(msg)

    return url


def validate_image_url(url: str) -> str:
    url_validator(url)

    extension = Path(url).suffix[1:]
    if not extension:
        msg = "Image URLs must have a file extension."
        raise ValidationError(msg)

    validate_is_allowed_image_extension(extension)

    return url


def validate_is_allowed_image_extension(extension: str) -> str:
    allowed_extensions = get_available_image_extensions()

    if extension.lower() not in allowed_extensions:
        allowed_str = comma_sep_str(allowed_extensions, last_sep="or", quote=True)
        msg = f"File extension '{extension}' is not allowed. Allowed extensions are: {allowed_str}."
        raise ValidationError(msg)

    return extension


def validate_file_upload(value: UploadedFile) -> UploadedFile:
    if not value.name:
        msg = "No filename could be determined."
        raise ValidationError(msg)

    return value


def validate_image_upload(value: UploadedFile) -> UploadedFile:
    try:
        from PIL import Image  # noqa: PLC0415
    except ImportError as error:  # pragma: no cover
        msg = "The `Pillow` library is not installed. Could not perform image validation."
        raise ValidationError(msg) from error

    if not value.name:
        msg = "No filename could be determined."
        raise ValidationError(msg)

    try:
        name, extension = value.name.rsplit(".", 1)
    except ValueError as error:
        msg = "Filename must have two parts: the name and the extension."
        raise ValidationError(msg) from error

    if not name:
        msg = "Filename must not be empty."
        raise ValidationError(msg)

    if not extension:
        msg = "Filename must have an extension."
        raise ValidationError(msg)

    extension = value.name.split(".")[-1]
    validate_is_allowed_image_extension(extension)

    file = BytesIO(value.read())
    value.seek(0)  # Go back to the beginning of the file after reading it

    try:
        Image.open(file).verify()
    except Exception as error:
        msg = "File either not an image or a corrupted image."
        raise ValidationError(msg) from error

    return value



from __future__ import annotations

import datetime
import math
import uuid
from typing import Any

from graphql.type.scalars import GRAPHQL_MAX_INT, GRAPHQL_MIN_INT

from ._definition import ScalarType

__all__ = [
    "GraphQLAny",
    "any_scalar",
]


any_scalar: ScalarType[Any, Any] = ScalarType(
    name="Any",
    description="Represent any value accepted by GraphQL.",
)

GraphQLAny = any_scalar.as_graphql_scalar()


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: str | bool | None) -> Any:  # noqa: FBT001
    return value


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: bytes) -> Any:
    return value.decode(encoding="utf-8")


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: int) -> Any:
    if not (GRAPHQL_MIN_INT <= value <= GRAPHQL_MAX_INT):
        msg = "GraphQL integers cannot represent non 32-bit signed integer value."
        raise ValueError(msg)
    return value


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: float) -> Any:
    if not math.isfinite(value):
        msg = "GraphQL floats cannot represent 'inf' or 'NaN' values."
        raise ValueError(msg)
    return value


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: datetime.datetime | datetime.date | datetime.time) -> Any:
    return value.isoformat()


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: uuid.UUID) -> Any:
    return str(value)


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: list) -> Any:
    return [any_scalar.parse(value) for value in value]


@any_scalar.serialize.register
@any_scalar.parse.register
def _(value: dict) -> Any:
    return {key: any_scalar.parse(value) for key, value in value.items()}



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from collections.abc import Generator

    from pytest_undine.client import AsyncGraphQLClient, GraphQLClient
    from undine.settings import UndineDefaultSettings

__all__ = [
    "graphql",
    "graphql_async",
    "undine_settings",
]


@pytest.fixture
def graphql() -> GraphQLClient:
    from .client import GraphQLClient  # noqa: PLC0415

    return GraphQLClient()


@pytest.fixture
def graphql_async() -> AsyncGraphQLClient:
    from .client import AsyncGraphQLClient  # noqa: PLC0415

    return AsyncGraphQLClient()


@pytest.fixture
def undine_settings() -> Generator[UndineDefaultSettings, None, None]:
    from undine.settings import undine_settings  # noqa: PLC0415

    try:
        yield undine_settings
    finally:
        undine_settings.reload()  # type: ignore[attr-defined]



from __future__ import annotations

from collections.abc import Callable, Iterable
from functools import wraps
from typing import Any, Generic, NoReturn, TypeVar

from django.core.exceptions import ValidationError
from graphql import DirectiveLocation, GraphQLScalarType
from graphql.pyutils import inspect

from undine.directives import Directive
from undine.exceptions import GraphQLScalarConversionError, GraphQLScalarTypeNotSupportedError
from undine.settings import undine_settings
from undine.typing import DispatchProtocol, T
from undine.utils.function_dispatcher import FunctionDispatcher
from undine.utils.graphql.type_registry import get_or_create_graphql_scalar
from undine.utils.graphql.utils import check_directives
from undine.utils.text import dotpath

__all__ = [
    "ScalarType",
]

TParse = TypeVar("TParse")
TSerialize = TypeVar("TSerialize")


class ScalarType(Generic[TParse, TSerialize]):
    """
    Create a new scalar for the GraphQL Schema.

    >>> my_scalar = ScalarType(name="MyScalar")
    >>> my_scalar.as_graphql_scalar()
    <GraphQLScalarType "MyScalar">
    """

    def __init__(
        self,
        name: str,
        description: str | None = None,
        specified_by_url: str | None = None,
        directives: Iterable[Directive] | None = None,
        extensions: dict[str, Any] | None = None,
    ) -> None:
        """
        Create a new `ScalarType`.

        :param name: The name of the 'ScalarType'.
        :param description: Description of the 'ScalarType'.
        :param specified_by_url: URL to the specification of the 'ScalarType'.
        :param directives: GraphQL directives for the 'ScalarType'.
        :param extensions: GraphQL extensions for the 'ScalarType'.
        """
        self.name = name
        self.description = description
        self.specified_by_url = specified_by_url
        self.directives = directives or []
        self.extensions = extensions or {}

        check_directives(self.directives, location=DirectiveLocation.SCALAR)
        self.extensions[undine_settings.SCALAR_EXTENSIONS_KEY] = self

        error_wrapper = handle_scalar_errors(name)
        self.parse: FunctionDispatcher[TParse] = FunctionDispatcher(wrapper=error_wrapper)
        self.serialize: FunctionDispatcher[TSerialize] = FunctionDispatcher(wrapper=error_wrapper)

        @self.serialize.register
        @self.parse.register
        def _(value: Any) -> NoReturn:
            raise GraphQLScalarTypeNotSupportedError(input_type=type(value))

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(name={self.name!r})>"

    def __str__(self) -> str:
        return undine_settings.SDL_PRINTER.print_scalar_type(self.as_graphql_scalar())

    def as_graphql_scalar(self) -> GraphQLScalarType:
        return get_or_create_graphql_scalar(
            name=self.name,
            description=self.description,
            serialize=self.serialize,
            parse_value=self.parse,
            specified_by_url=self.specified_by_url,
            extensions=self.extensions,
        )

    def __add_directive__(self, directive: Directive, /) -> ScalarType:
        """Add a directive to this scalar."""
        check_directives([directive], location=DirectiveLocation.SCALAR)
        self.directives.append(directive)
        return self


def handle_scalar_errors(typename: str) -> Callable[[DispatchProtocol[T]], DispatchProtocol[T]]:
    """Catch errors raised by the scalar parsers and serializers and reraise them as GraphQLErrors."""

    def decorator(func: DispatchProtocol[T]) -> DispatchProtocol[T]:
        @wraps(func)
        def wrapper(key: Any, **kwargs: Any) -> T:
            try:
                return func(key)
            except GraphQLScalarConversionError:
                raise
            except ValidationError as err:
                msg = str(err.message % err.params) if err.params else str(err.message)
                raise GraphQLScalarConversionError(typename=typename, value=inspect(key), error=msg) from err
            except Exception as err:
                raise GraphQLScalarConversionError(typename=typename, value=inspect(key), error=str(err)) from err

        return wrapper  # type: ignore[return-value]

    return decorator



from __future__ import annotations

import functools
import itertools
import operator
from collections import defaultdict
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Self, Unpack

from django.db.models import OrderBy
from graphql import DirectiveLocation, GraphQLEnumValue, Undefined

from undine.converters import convert_to_description, convert_to_graphql_type, convert_to_order_ref
from undine.dataclasses import OrderResults
from undine.exceptions import (
    GraphQLInvalidOrderDataError,
    MismatchingModelError,
    MissingModelGenericError,
    NotCompatibleWithError,
    QueryTypeRequiresSingleModelError,
    UnionTypeModelsDifferentError,
    UnionTypeRequiresMultipleModelsError,
)
from undine.parsers import parse_class_attribute_docstrings
from undine.settings import undine_settings
from undine.typing import TModels
from undine.utils.graphql.type_registry import get_or_create_graphql_enum
from undine.utils.graphql.utils import check_directives
from undine.utils.model_utils import get_model_field, get_model_fields_for_graphql
from undine.utils.reflection import get_members, get_wrapped_func, is_subclass
from undine.utils.text import dotpath, get_docstring, to_schema_name

if TYPE_CHECKING:
    from collections.abc import Iterable

    from django.db.models import Model
    from graphql import GraphQLEnumType, GraphQLInputType

    from undine import QueryType, UnionType
    from undine.directives import Directive
    from undine.typing import (
        DjangoExpression,
        DjangoRequestProtocol,
        GQLInfo,
        OrderAliasesFunc,
        OrderParams,
        OrderSetParams,
        T,
        TModel,
        VisibilityFunc,
    )

__all__ = [
    "Order",
    "OrderSet",
]


class OrderSetMeta(type):
    """A metaclass that modifies how an `OrderSet` is created."""

    # Set in '__new__'
    __models__: tuple[type[Model], ...]
    __order_map__: dict[str, Order]
    __schema_name__: str
    __directives__: list[Directive]
    __extensions__: dict[str, Any]
    __attribute_docstrings__: dict[str, str]

    def __new__(
        cls,
        _name: str,
        _bases: tuple[type, ...],
        _attrs: dict[str, Any],
        **kwargs: Unpack[OrderSetParams],
    ) -> OrderSetMeta:
        if _name == "OrderSet":  # Early return for the `OrderSet` class itself.
            return super().__new__(cls, _name, _bases, _attrs)

        try:
            models = OrderSetMeta.__models__
            del OrderSetMeta.__models__
        except AttributeError as error:
            raise MissingModelGenericError(name=_name, cls="OrderSet") from error

        auto = kwargs.get("auto", undine_settings.AUTOGENERATION)
        exclude = set(kwargs.get("exclude", []))
        if auto:
            exclude |= set(_attrs)

            if len(models) == 1:
                _attrs |= get_orders_for_model(models[0], exclude=exclude)
            else:
                _attrs |= get_orders_for_models(models, exclude=exclude)

        orderset = super().__new__(cls, _name, _bases, _attrs)

        # Members should use `__dunder__` names to avoid name collisions with possible `Order` names.
        orderset.__models__ = models
        orderset.__order_map__ = get_members(orderset, Order)
        orderset.__schema_name__ = kwargs.get("schema_name", _name)
        orderset.__directives__ = kwargs.get("directives", [])
        orderset.__extensions__ = kwargs.get("extensions", {})
        orderset.__attribute_docstrings__ = parse_class_attribute_docstrings(orderset)

        check_directives(orderset.__directives__, location=DirectiveLocation.ENUM)
        orderset.__extensions__[undine_settings.ORDERSET_EXTENSIONS_KEY] = orderset

        for name, order in orderset.__order_map__.items():
            order.__connect__(orderset, name)  # type: ignore[arg-type]

        return orderset

    def __str__(cls) -> str:
        return undine_settings.SDL_PRINTER.print_enum_type(cls.__enum_type__())

    def __getitem__(cls, models: type[type[TModel]] | tuple[type[TModel], ...]) -> type[OrderSet[*TModels]]:
        # Note that this should be cleaned up in '__new__',
        # but is not if an error occurs in the class body of the defined 'OrderSet'!
        OrderSetMeta.__models__ = models if isinstance(models, tuple) else (models,)
        return cls  # type: ignore[return-value]

    def __call__(cls, ref: T) -> T:
        """
        Allow adding this OrderSet to a QueryType using a decorator syntax

        >>> class TaskOrderSet(OrderSet[Task]): ...
        >>>
        >>> @TaskOrderSet
        >>> class TaskType(QueryType[Task]): ...
        """
        from undine import QueryType, UnionType  # noqa: PLC0415

        if is_subclass(ref, QueryType):
            cls.__add_to_query_type__(ref)
        elif is_subclass(ref, UnionType):
            cls.__add_to_union_type__(ref)
        else:
            raise NotCompatibleWithError(obj=cls, other=ref)

        return ref

    def __build__(cls, order_data: list[str], info: GQLInfo) -> OrderResults:
        """
        Build a list of 'OrderBy' expressions from the given order input data.

        :param order_data: The input order data.
        :param info: The GraphQL resolve info for the request.
        """
        order_by: list[OrderBy] = []
        aliases: dict[str, DjangoExpression] = {}
        order_count: int = 0

        for enum_value in order_data:
            if enum_value.endswith("_desc"):
                order_name = enum_value.removesuffix("_desc")
                descending = True
            elif enum_value.endswith("_asc"):
                order_name = enum_value.removesuffix("_asc")
                descending = False
            else:  # pragma: no cover
                raise GraphQLInvalidOrderDataError(orderset=cls, enum_value=enum_value)

            order = cls.__order_map__[order_name]
            if order.aliases_func is not None:
                aliases |= order.aliases_func(order, info, descending=descending)

            expression = order.get_expression(descending=descending)
            order_by.append(expression)
            order_count += 1

        return OrderResults(order_by=order_by, aliases=aliases, order_count=order_count)

    def __enum_type__(cls) -> GraphQLEnumType:
        """Create the enum type to use for the `QueryType` this `OrderSet` is for."""
        return get_or_create_graphql_enum(
            name=cls.__schema_name__,
            values=cls.__enum_values__(),
            description=get_docstring(cls),
            extensions=cls.__extensions__,
        )

    def __enum_values__(cls) -> dict[str, GraphQLEnumValue | str]:
        """
        Get all the enum values for this `OrderSet`.
        The values are the names of all the `Order` instances defined on this `OrderSet`,
        in both ascending and descending directions.
        """
        enum_values: dict[str, GraphQLEnumValue | str] = {}

        for ordering in cls.__order_map__.values():
            for descending in (False, True):
                name = ordering.schema_name + ("Desc" if descending else "Asc")
                enum_value = ordering.as_graphql_enum_value()
                enum_value.value += "_desc" if descending else "_asc"
                enum_values[name] = enum_value

        return enum_values

    def __add_directive__(cls, directive: Directive, /) -> Self:
        """Add a directive to this orderset."""
        check_directives([directive], location=DirectiveLocation.ENUM)
        cls.__directives__.append(directive)
        return cls

    def __add_to_query_type__(cls, query_type: type[QueryType]) -> None:
        models = cls.__models__
        if len(models) != 1:
            raise QueryTypeRequiresSingleModelError(kind="OrderSet")

        if models[0] is not query_type.__model__:
            raise MismatchingModelError(
                name=cls.__name__,
                given_model=models[0],
                target=query_type.__name__,
                expected_model=query_type.__model__,
            )

        query_type.__orderset__ = cls  # type: ignore[assignment]

    def __add_to_union_type__(cls, union_type: type[UnionType]) -> None:
        models = cls.__models__
        if len(models) == 1:
            raise UnionTypeRequiresMultipleModelsError(kind="OrderSet")

        on_union_type = set(union_type.__query_types_by_model__)
        on_orderset = set(models)

        if on_union_type != on_orderset:
            raise UnionTypeModelsDifferentError(kind="OrderSet")

        union_type.__orderset__ = cls  # type: ignore[assignment]


class OrderSet(Generic[*TModels], metaclass=OrderSetMeta):
    """
    A class for adding ordering for a `QueryType`.

    Must set the Django Model this `OrderSet` is for using the generic type argument.
    Model must match the Model of the `QueryType` this `OrderSet` will be added to.

    The following parameters can be passed in the class definition:

    `auto: bool = <AUTOGENERATION setting>`
        Whether to add `Order` attributes for all Model fields automatically.

    `exclude: list[str] = []`
        Model fields to exclude from automatically added `Order` attributes.

    `schema_name: str = <class name>`
        Override the name for the `GraphQLEnum` for this `OrderSet` in the GraphQL schema.

    `directives: list[Directive] = []`
        `Directives` to add to the created `GraphQLEnum`.

    `extensions: dict[str, Any] = {}`
        GraphQL extensions for the created `GraphQLEnum`.

    >>> class TaskOrderSet(OrderSet[Task]): ...
    >>> class TaskQueryType(QueryType[Task], orderset=TaskOrderSet): ...
    """

    # Members should use `__dunder__` names to avoid name collisions with possible `Order` names.

    # Set in metaclass
    __models__: ClassVar[tuple[type[Model], ...]]
    __order_map__: ClassVar[dict[str, Order]]
    __schema_name__: ClassVar[str]
    __directives__: ClassVar[list[Directive]]
    __extensions__: ClassVar[dict[str, Any]]
    __attribute_docstrings__: ClassVar[dict[str, str]]

    @classmethod
    def __is_visible__(cls, request: DjangoRequestProtocol) -> bool:
        """
        Determine if the given `OrderSet` is visible in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.
        """
        return True


class Order:
    """
    A class for defining a possible ordering for a QueryType.
    Represents a value in the GraphQL `EnumType` for the `OrderSet` this is added to.

    >>> class TaskOrderSet(OrderSet[Task]):
    ...     name = Order()
    """

    def __init__(self, ref: Any = None, **kwargs: Unpack[OrderParams]) -> None:
        """
        Create a new `Order`.

        :param ref: the expression to order by. Must be convertable by the `convert_to_order_ref` function.
                    If not provided, use the name of the attribute this is assigned to in the `OrderSet` class.
        :param null_placement: Where should null values be placed? By default, use database default.
        :param description: Description of the `Order`.
        :param deprecation_reason: If this `Order` is deprecated, describes the reason for deprecation.
        :param field_name: Name of the field in the Django model. If not provided, use the name of the attribute.
        :param schema_name: Actual name of the `Order` in the GraphQL schema. Can be used to alias the `Order`
                            for the schema, or when the desired name is a Python keyword (e.g. `if` or `from`).
        :param directives: GraphQL directives for the `Order`.
        :param extensions: GraphQL extensions for the `Order`.
        """
        self.ref: Any = ref

        self.nulls_first: bool | None = True if kwargs.get("null_placement") == "first" else None
        self.nulls_last: bool | None = True if kwargs.get("null_placement") == "last" else None
        self.description: str | None = kwargs.get("description", Undefined)  # type: ignore[assignment]
        self.deprecation_reason: str | None = kwargs.get("deprecation_reason")
        self.field_name: str = kwargs.get("field_name", Undefined)  # type: ignore[assignment]
        self.schema_name: str = kwargs.get("schema_name", Undefined)  # type: ignore[assignment]
        self.directives: list[Directive] = kwargs.get("directives", [])
        self.extensions: dict[str, Any] = kwargs.get("extensions", {})

        check_directives(self.directives, location=DirectiveLocation.ENUM_VALUE)
        self.extensions[undine_settings.ORDER_EXTENSIONS_KEY] = self

        self.aliases_func: OrderAliasesFunc | None = None
        self.visible_func: VisibilityFunc | None = None

    def __connect__(self, orderset: type[OrderSet], name: str) -> None:
        """Connect this `Order` to the given `OrderSet` using the given name."""
        self.orderset = orderset
        self.name = name
        self.field_name = self.field_name or name
        self.schema_name = self.schema_name or to_schema_name(name)

        if isinstance(self.ref, str):
            self.field_name = self.ref

        self.ref = convert_to_order_ref(self.ref, caller=self)

        if self.description is Undefined:
            self.description = self.orderset.__attribute_docstrings__.get(name)
            if self.description is None:
                self.description = convert_to_description(self.ref)

    def __repr__(self) -> str:
        return f"<{dotpath(self.__class__)}(ref={self.ref!r})>"

    def __str__(self) -> str:
        value = self.as_graphql_enum_value()
        return undine_settings.SDL_PRINTER.print_enum_value(self.schema_name, value, indent=False)

    def get_expression(self, *, descending: bool) -> OrderBy:
        return OrderBy(
            expression=self.ref,
            nulls_first=self.nulls_first,
            nulls_last=self.nulls_last,
            descending=descending,
        )

    def as_graphql_enum_value(self) -> GraphQLEnumValue:
        return GraphQLEnumValue(
            value=self.name,
            description=self.description,
            deprecation_reason=self.deprecation_reason,
            extensions=self.extensions,
        )

    def aliases(self, func: OrderAliasesFunc | None = None, /) -> OrderAliasesFunc:
        """
        Decorate a function to add additional queryset aliases required by this Order.

        >>> class TaskOrderSet(OrderSet[Task]):
        ...     name = Order()
        ...
        ...     @name.aliases
        ...     def name_aliases(self: Order, info: GQLInfo, *, value: str) -> dict[str, DjangoExpression]:
        ...         return {"foo": Value("bar")}
        """
        if func is None:  # Allow `@<order_name>.aliases()`
            return self.aliases  # type: ignore[return-value]
        self.aliases_func = get_wrapped_func(func)
        return func

    def visible(self, func: VisibilityFunc | None = None, /) -> VisibilityFunc:
        """
        Decorate a function to change the Order's visibility in the schema.
        Experimental, requires `EXPERIMENTAL_VISIBILITY_CHECKS` to be enabled.

        >>> class TaskOrderSet(OrderSet[Task]):
        ...     name = Order()
        ...
        ...     @name.visible
        ...     def name_visible(self: Order, request: DjangoRequestProtocol) -> bool:
        ...         return False
        """
        if func is None:  # Allow `@<order_name>.visible()`
            return self.visible  # type: ignore[return-value]
        self.visible_func = get_wrapped_func(func)
        return func

    def add_directive(self, directive: Directive, /) -> Self:
        """Add a directive to this order."""
        check_directives([directive], location=DirectiveLocation.ENUM_VALUE)
        self.directives.append(directive)
        return self


def get_orders_for_model(model: type[Model], *, exclude: Iterable[str] = ()) -> dict[str, Order]:
    """Creates `Orders` for all the given model's fields, except those in the 'exclude' list."""
    result: dict[str, Order] = {}

    for model_field in get_model_fields_for_graphql(model):
        field_name = model_field.name

        is_primary_key = bool(getattr(model_field, "primary_key", False))
        if is_primary_key:
            field_name = "pk"

        if field_name in exclude:
            continue

        result[field_name] = Order(field_name)

    return result


def get_orders_for_models(models: tuple[type[Model], ...], exclude: Iterable[str] = ()) -> dict[str, Order]:
    result: dict[str, Order] = {}

    fields_by_model: dict[type[Model], set[str]] = {}
    for model in models:
        fields: set[str] = set()

        for model_field in get_model_fields_for_graphql(model):
            field_name = model_field.name

            is_primary_key = bool(getattr(model_field, "primary_key", False))
            if is_primary_key:
                field_name = "pk"

            if field_name in exclude:
                continue

            fields.add(field_name)

        fields_by_model[model] = fields

    common_fields = functools.reduce(operator.and_, fields_by_model.values())

    graphql_types_by_model: dict[str, dict[type[Model], GraphQLInputType]] = defaultdict(dict)
    for model in fields_by_model:
        for field_name in common_fields:
            graphql_types_by_model[field_name][model] = convert_to_graphql_type(field_name, model=model, is_input=True)

    usable_fields: set[str] = set()
    for field_name, model_map in graphql_types_by_model.items():
        is_usable = all(field_1 == field_2 for field_1, field_2 in itertools.combinations(model_map.values(), 2))
        if is_usable:
            usable_fields.add(field_name)

    for field_name in usable_fields:
        model_field = get_model_field(model=models[0], lookup=field_name)

        is_primary_key = bool(getattr(model_field, "primary_key", False))
        if is_primary_key:
            field_name = "pk"  # noqa: PLW2901

        if field_name in exclude:
            continue

        result[field_name] = Order(field_name)

    return result



from __future__ import annotations

import textwrap
from itertools import chain, starmap
from typing import TYPE_CHECKING

from graphql import (
    DEFAULT_DEPRECATION_REASON,
    GraphQLEnumType,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLScalarType,
    GraphQLUnionType,
    ast_from_value,
    is_specified_directive,
    print_ast,
)
from graphql.language.block_string import is_printable_as_block_string, print_block_string
from graphql.language.print_string import print_string
from graphql.pyutils import inspect
from graphql.utilities.print_schema import is_defined_type

from undine.utils.graphql.undine_extensions import (
    get_undine_calculation_argument,
    get_undine_directive_argument,
    get_undine_entrypoint,
    get_undine_field,
    get_undine_filter,
    get_undine_filterset,
    get_undine_input,
    get_undine_interface_field,
    get_undine_interface_type,
    get_undine_mutation_type,
    get_undine_order,
    get_undine_orderset,
    get_undine_query_type,
    get_undine_root_type,
    get_undine_scalar,
    get_undine_schema_directives,
    get_undine_union_type,
)
from undine.utils.graphql.validation_rules.one_of_input_object import is_one_of_input_object

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from graphql import (
        GraphQLArgument,
        GraphQLDirective,
        GraphQLEnumValue,
        GraphQLField,
        GraphQLInputField,
        GraphQLNamedType,
        GraphQLSchema,
    )

    from undine.directives import Directive

__all__ = [
    "SDLPrinter",
]


class SDLPrinter:  # noqa: PLR0904
    """Print the given GraphQL schema to GraphQL SDL."""

    @classmethod
    def print_schema(
        cls,
        schema: GraphQLSchema,
        *,
        directive_filter: Callable[[GraphQLDirective], bool] | None = None,
        type_filter: Callable[[GraphQLNamedType], bool] | None = None,
    ) -> str:
        if directive_filter is None:
            directive_filter = cls.default_directive_filter
        if type_filter is None:
            type_filter = cls.default_type_filter

        schema_definition = cls.print_schema_definition(schema)

        directives = (
            cls.print_directive(directive)  # ...
            for directive in schema.directives
            if directive_filter(directive)
        )
        types = (
            cls.print_type(type_)  # ...
            for type_ in schema.type_map.values()
            if type_filter(type_)
        )

        return "\n\n".join(
            chain(
                directives,
                types,
                [schema_definition] if schema_definition else [],
            )
        )

    @classmethod
    def print_schema_definition(cls, schema: GraphQLSchema) -> str:
        root_types: dict[str, str] = {}
        description = cls.print_docstring(schema.description)

        schema_str: str = "schema"
        has_directives = False

        undine_schema_directives = get_undine_schema_directives(schema)
        if undine_schema_directives is not None:
            for directive in undine_schema_directives:
                has_directives = True
                schema_str += cls.print_directive_usage(directive)

        if schema.query_type is not None:
            root_types[schema.query_type.name] = f"  query: {schema.query_type.name}"

        if schema.mutation_type is not None:
            root_types[schema.mutation_type.name] = f"  mutation: {schema.mutation_type.name}"

        if schema.subscription_type is not None:
            root_types[schema.subscription_type.name] = f"  subscription: {schema.subscription_type.name}"

        non_default_root_types = set(root_types) - {"Query", "Mutation", "Subscription"}
        has_default_root_types = not non_default_root_types

        # The schema definition only needs to be printed if its not the default schema definition.
        if not description and (not root_types or has_default_root_types) and not has_directives:
            return ""

        schema_str += cls.print_block(root_types.values())
        if description:
            schema_str = f"{description}\n{schema_str}"

        return schema_str

    @classmethod
    def print_type(cls, named_type: GraphQLNamedType) -> str:
        match named_type:
            case GraphQLScalarType():
                return cls.print_scalar_type(named_type)

            case GraphQLObjectType():
                return cls.print_object_type(named_type)

            case GraphQLInterfaceType():
                return cls.print_interface_type(named_type)

            case GraphQLUnionType():
                return cls.print_union_type(named_type)

            case GraphQLEnumType():
                return cls.print_enum_type(named_type)

            case GraphQLInputObjectType():
                return cls.print_input_object_type(named_type)

            case _:
                msg = f"Unexpected type: {inspect(named_type)}."
                raise TypeError(msg)

    # --- ObjectType --------------------------------------------------------------------------------------

    @classmethod
    def print_object_type(cls, object_type: GraphQLObjectType) -> str:
        object_type_str = f"type {object_type.name}"

        if object_type.interfaces:
            object_type_str += " implements " + " & ".join(interface.name for interface in object_type.interfaces)

        undine_query_type = get_undine_query_type(object_type)
        if undine_query_type is not None:
            for directive in undine_query_type.__directives__:
                object_type_str += cls.print_directive_usage(directive)

        undine_root_type = get_undine_root_type(object_type)
        if undine_root_type is not None:
            for directive in undine_root_type.__directives__:
                object_type_str += cls.print_directive_usage(directive)

        if object_type.fields:
            fields = starmap(cls.print_field, object_type.fields.items())
            object_type_str += cls.print_block(fields)

        description = cls.print_docstring(object_type.description)
        if description:
            object_type_str = f"{description}\n{object_type_str}"

        return object_type_str

    @classmethod
    def print_interface_type(cls, interface_type: GraphQLInterfaceType) -> str:
        interface_type_str = f"interface {interface_type.name}"

        if interface_type.interfaces:
            interface_type_str += " implements " + " & ".join(interface.name for interface in interface_type.interfaces)

        undine_interface = get_undine_interface_type(interface_type)
        if undine_interface is not None:
            for directive in undine_interface.__directives__:
                interface_type_str += cls.print_directive_usage(directive)

        if interface_type.fields:
            fields = starmap(cls.print_field, interface_type.fields.items())
            interface_type_str += cls.print_block(fields)

        description = cls.print_docstring(interface_type.description)
        if description:
            interface_type_str = f"{description}\n{interface_type_str}"

        return interface_type_str

    @classmethod
    def print_field(cls, name: str, field: GraphQLField, *, indent: bool = True) -> str:
        indentation: str = "  " if indent else ""

        args_str: str = ""
        if field.args:
            args: Iterable[str] = starmap(cls.print_field_argument, field.args.items())
            if not indent:
                args = (textwrap.indent(textwrap.dedent(arg), prefix="  ") for arg in args)

            args_str = "(\n" + "\n".join(args) + f"\n{indentation})"

        field_str: str = f"{indentation}{name}{args_str}: {field.type}"

        field_str += cls.print_deprecated(field.deprecation_reason)

        undine_entrypoint = get_undine_entrypoint(field)
        if undine_entrypoint is not None:
            for directive in undine_entrypoint.directives:
                field_str += cls.print_directive_usage(directive)

        undine_field = get_undine_field(field)
        if undine_field is not None:
            if undine_field.complexity != 0:
                field_str += f" @complexity(value: {undine_field.complexity})"

            for directive in undine_field.directives:
                field_str += cls.print_directive_usage(directive)

        undine_interface_field = get_undine_interface_field(field)
        if undine_interface_field is not None:
            for directive in undine_interface_field.directives:
                field_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(field.description)
        if description:
            description = description.replace("\n", f"\n{indentation}").strip()
            field_str = f"{indentation}{description}\n{field_str}"

        return field_str

    @classmethod
    def print_field_argument(cls, name: str, arg: GraphQLArgument, *, indent: bool = True) -> str:
        indentation: str = "    " if indent else ""

        arg_str: str = f"{indentation}{name}: {arg.type}"

        default_ast = ast_from_value(arg.default_value, arg.type)
        if default_ast:
            arg_str += f" = {print_ast(default_ast)}"

        arg_str += cls.print_deprecated(arg.deprecation_reason)

        undine_calculation_argument = get_undine_calculation_argument(arg)
        if undine_calculation_argument is not None:
            for directive in undine_calculation_argument.directives:
                arg_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(arg.description)
        if description:
            description = description.replace("\n", f"\n{indentation}").strip()
            arg_str = f"{indentation}{description}\n{arg_str}"

        return arg_str

    # --- InputObjectType ---------------------------------------------------------------------------------

    @classmethod
    def print_input_object_type(cls, input_object_type: GraphQLInputObjectType) -> str:
        input_object_type_str = f"input {input_object_type.name}"

        if is_one_of_input_object(input_object_type):
            input_object_type_str += " @oneOf"

        undine_mutation_type = get_undine_mutation_type(input_object_type)
        if undine_mutation_type is not None:
            for directive in undine_mutation_type.__directives__:
                input_object_type_str += cls.print_directive_usage(directive)

        undine_filterset = get_undine_filterset(input_object_type)
        if undine_filterset is not None:
            for directive in undine_filterset.__directives__:
                input_object_type_str += cls.print_directive_usage(directive)

        if input_object_type.fields:
            input_fields = starmap(cls.print_input_field, input_object_type.fields.items())
            input_object_type_str += cls.print_block(input_fields)

        description = cls.print_docstring(input_object_type.description)
        if description:
            input_object_type_str = f"{description}\n{input_object_type_str}"

        return input_object_type_str

    @classmethod
    def print_input_field(cls, name: str, input_field: GraphQLInputField, *, indent: bool = True) -> str:
        indentation: str = "  " if indent else ""

        input_field_str: str = f"{indentation}{name}: {input_field.type}"

        default_ast = ast_from_value(input_field.default_value, input_field.type)
        if default_ast:
            input_field_str += f" = {print_ast(default_ast)}"

        input_field_str += cls.print_deprecated(input_field.deprecation_reason)

        undine_input = get_undine_input(input_field)
        if undine_input is not None:
            for directive in undine_input.directives:
                input_field_str += cls.print_directive_usage(directive)

        undine_filter = get_undine_filter(input_field)
        if undine_filter is not None:
            for directive in undine_filter.directives:
                input_field_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(input_field.description)
        if description:
            description = description.replace("\n", f"\n{indentation}").strip()
            input_field_str = f"{indentation}{description}\n{input_field_str}"

        return input_field_str

    # --- EnumType ----------------------------------------------------------------------------------------

    @classmethod
    def print_enum_type(cls, enum_type: GraphQLEnumType) -> str:
        enum_str: str = f"enum {enum_type.name}"

        undine_orderset = get_undine_orderset(enum_type)
        if undine_orderset is not None:
            for directive in undine_orderset.__directives__:
                enum_str += cls.print_directive_usage(directive)

        if enum_type.values:
            enum_values = starmap(cls.print_enum_value, enum_type.values.items())
            enum_str += cls.print_block(enum_values)

        description = cls.print_docstring(enum_type.description)
        if description:
            enum_str = f"{description}\n{enum_str}"

        return enum_str

    @classmethod
    def print_enum_value(cls, name: str, value: GraphQLEnumValue, *, indent: bool = True) -> str:
        indentation: str = "  " if indent else ""

        enum_value_str: str = f"{indentation}{name}"
        enum_value_str += cls.print_deprecated(value.deprecation_reason)

        undine_order = get_undine_order(value)
        if undine_order is not None:
            for directive in undine_order.directives:
                enum_value_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(value.description)
        if description:
            description = description.replace("\n", f"\n{indentation}").strip()
            enum_value_str = f"{indentation}{description}\n{enum_value_str}"

        return enum_value_str

    # --- ScalarType --------------------------------------------------------------------------------------

    @classmethod
    def print_scalar_type(cls, scalar_type: GraphQLScalarType) -> str:
        scalar_str = f"scalar {scalar_type.name}"

        if scalar_type.specified_by_url is not None:
            scalar_str += f" @specifiedBy(url: {print_string(scalar_type.specified_by_url)})"

        undine_scalar = get_undine_scalar(scalar_type)
        if undine_scalar is not None:
            for directive in undine_scalar.directives:
                scalar_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(scalar_type.description)
        if description:
            scalar_str = f"{description}\n{scalar_str}"

        return scalar_str

    # --- UnionType ---------------------------------------------------------------------------------------

    @classmethod
    def print_union_type(cls, union_type: GraphQLUnionType) -> str:
        union_str = f"union {union_type.name}"

        undine_union = get_undine_union_type(union_type)
        if undine_union is not None:
            for directive in undine_union.__directives__:
                union_str += cls.print_directive_usage(directive)

        if union_type.types:
            possible_types = " | ".join(type_.name for type_ in union_type.types)
            union_str += f" = {possible_types}"

        description = cls.print_docstring(union_type.description)
        if description:
            union_str = f"{description}\n{union_str}"

        return union_str

    # --- Directives --------------------------------------------------------------------------------------

    @classmethod
    def print_directive(cls, directive: GraphQLDirective) -> str:
        directive_str = f"directive @{directive.name}"

        if directive.args:
            args = starmap(cls.print_directive_argument, directive.args.items())
            directive_str += "(\n" + "\n".join(args) + "\n)"

        if directive.is_repeatable:
            directive_str += " repeatable"

        if directive.locations:
            directive_str += " on "
            directive_str += " | ".join(location.name for location in directive.locations)

        description = cls.print_docstring(directive.description)
        if description:
            directive_str = f"{description}\n{directive_str}"

        return directive_str

    @classmethod
    def print_directive_argument(cls, name: str, arg: GraphQLArgument, *, indent: bool = True) -> str:
        indentation: str = "  " if indent else ""

        arg_str: str = f"{indentation}{name}: {arg.type}"

        default_ast = ast_from_value(arg.default_value, arg.type)
        if default_ast:
            arg_str += f" = {print_ast(default_ast)}"

        arg_str += cls.print_deprecated(arg.deprecation_reason)

        undine_dir_arg = get_undine_directive_argument(arg)
        if undine_dir_arg is not None:
            for directive in undine_dir_arg.directives:
                arg_str += cls.print_directive_usage(directive)

        description = cls.print_docstring(arg.description)
        if description:
            description = description.replace("\n", f"\n{indentation}").strip()
            arg_str = f"{indentation}{description}\n{arg_str}"

        return arg_str

    @classmethod
    def print_directive_usage(cls, directive: Directive, *, indent: bool = True) -> str:
        indentation: str = " " if indent else ""

        output = f"{indentation}@{directive.__schema_name__}"
        if not directive.__parameters__:
            return output

        args: list[str] = []
        for parameter, value in directive.__parameters__.items():
            arg = directive.__arguments__[parameter]
            value_ast = ast_from_value(value, arg.input_type)
            if value_ast is None:
                continue

            arg_str = f"{arg.schema_name}: {print_ast(value_ast)}"
            args.append(arg_str)

        if not args:
            return output

        return f"{output}(" + ", ".join(args) + ")"

    @classmethod
    def print_deprecated(cls, reason: str | None, *, indent: bool = True) -> str:
        indentation: str = " " if indent else ""

        if reason is None:
            return ""
        if reason != DEFAULT_DEPRECATION_REASON:
            return f"{indentation}@deprecated(reason: {print_string(reason)})"
        return f"{indentation}@deprecated"

    # --- Utils --------------------------------------------------------------------------------------------

    @classmethod
    def print_docstring(cls, description: str | None) -> str:
        if description is None:
            return ""
        if is_printable_as_block_string(description):
            return print_block_string(description)
        return print_string(description)

    @classmethod
    def print_block(cls, values: Iterable[str]) -> str:
        return " {\n" + "\n".join(values) + "\n}"

    @classmethod
    def default_directive_filter(cls, directive: GraphQLDirective) -> bool:
        return not is_specified_directive(directive)

    @classmethod
    def default_type_filter(cls, named_type: GraphQLNamedType) -> bool:
        return is_defined_type(named_type)



from __future__ import annotations

from typing import Any, Literal

from django.db.models.constants import LOOKUP_SEP
from graphql import (
    GraphQLBoolean,
    GraphQLInputType,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLOutputType,
    GraphQLString,
)

from undine.converters import convert_lookup_to_graphql_type
from undine.exceptions import FunctionDispatcherError
from undine.scalars import GraphQLDate, GraphQLTime


@convert_lookup_to_graphql_type.register
def _(lookup: str, **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if LOOKUP_SEP not in lookup:
        msg = f"Could not find a matching GraphQL type for lookup: '{lookup}'."
        raise FunctionDispatcherError(msg)

    transform, rest = lookup.split(LOOKUP_SEP, maxsplit=1)
    kwargs["default_type"] = convert_lookup_to_graphql_type(transform, **kwargs)
    return convert_lookup_to_graphql_type(rest, **kwargs)


@convert_lookup_to_graphql_type.register
def _(_: Literal["exact"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return kwargs["default_type"]


@convert_lookup_to_graphql_type.register
def _(_: Literal["endswith", "startswith"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return kwargs["default_type"]


@convert_lookup_to_graphql_type.register
def _(_: Literal["contains"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return kwargs["default_type"]


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "icontains",
        "iendswith",
        "iexact",
        "iregex",
        "istartswith",
        "regex",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "gt",
        "gte",
        "lt",
        "lte",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return kwargs["default_type"]


@convert_lookup_to_graphql_type.register
def _(_: Literal["isnull"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBoolean


@convert_lookup_to_graphql_type.register
def _(_: Literal["in", "range"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    if isinstance(kwargs["default_type"], GraphQLList):
        return kwargs["default_type"]
    return GraphQLList(GraphQLNonNull(kwargs["default_type"]))


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "day",
        "hour",
        "iso_week_day",
        "iso_year",
        "microsecond",
        "minute",
        "month",
        "quarter",
        "second",
        "week",
        "week_day",
        "year",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLInt


@convert_lookup_to_graphql_type.register
def _(_: Literal["date"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLDate


@convert_lookup_to_graphql_type.register
def _(_: Literal["time"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLTime


@convert_lookup_to_graphql_type.register
def _(_: Literal["contained_by", "overlap"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return kwargs["default_type"]


@convert_lookup_to_graphql_type.register
def _(_: Literal["len"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLInt


@convert_lookup_to_graphql_type.register
def _(_: Literal["has_key"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "has_any_keys",
        "has_keys",
        "keys",
        "values",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLList(GraphQLNonNull(GraphQLString))


@convert_lookup_to_graphql_type.register
def _(_: Literal["unaccent"], **kwargs: Any) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "trigram_similar",
        "trigram_word_similar",
        "trigram_strict_word_similar",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLString


@convert_lookup_to_graphql_type.register
def _(
    _: Literal[
        "isempty",
        "lower_inc",
        "lower_inf",
        "upper_inc",
        "upper_inf",
    ],
    **kwargs: Any,
) -> GraphQLInputType | GraphQLOutputType:
    return GraphQLBoolean



from __future__ import annotations

from .calculation import Calculation, CalculationArgument
from .dataloaders import DataLoader
from .entrypoint import Entrypoint, RootType
from .filtering import Filter, FilterSet
from .interface import InterfaceField, InterfaceType
from .mutation import Input, MutationType
from .ordering import Order, OrderSet
from .query import Field, QueryType
from .schema import create_schema
from .typing import DjangoExpression, GQLInfo
from .union import UnionType

__all__ = [
    "Calculation",
    "CalculationArgument",
    "DataLoader",
    "DjangoExpression",
    "Entrypoint",
    "Field",
    "Filter",
    "FilterSet",
    "GQLInfo",
    "Input",
    "InterfaceField",
    "InterfaceType",
    "MutationType",
    "Order",
    "OrderSet",
    "QueryType",
    "RootType",
    "UnionType",
    "create_schema",
]
